# Copyright (c) OpenMMLab. All rights reserved.

import torch
import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from mmengine.model.weight_init import trunc_normal_

from .swinir_modules import PatchEmbed, PatchUnEmbed
from .swinir_utils import (drop_path, to_2tuple, window_partition,
                           window_reverse)


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of
    residual blocks)."""

    def __init__(self, drop_prob: float = 0., scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).

        Returns:
            Tensor: Forward results.
        """
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f'drop_prob={round(self.drop_prob, 3):0.3f}'


class Mlp(nn.Module):
    """Multilayer Perceptron layer.

    Args:
        in_features (int): Number of input channels.
        hidden_features (int | None, optional): Number of hidden layer
            channels. Default: None
        out_features (int | None, optional): Number of output channels.
            Default: None
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        drop (float, optional): Dropout ratio of attention weight. Default: 0.0
    """

    def __init__(self,
                 in_features,
                 hidden_features=None,
                 out_features=None,
                 act_layer=nn.GELU,
                 drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).

        Returns:
            Tensor: Forward results.
        """
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class WindowAttention(nn.Module):
    r""" Window based multi-head self attention (W-MSA)
    module with relative position bias.
    It supports both of shifted and non-shifted window.
    Args:
        dim (int): Number of input channels.
        window_size (tuple[int]): The height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to
            query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale
            of head_dim ** -0.5 if set
        attn_drop (float, optional): Dropout ratio of attention weight.
            Default: 0.0
        proj_drop (float, optional): Dropout ratio of output. Default: 0.0
    """

    def __init__(self,
                 dim,
                 window_size,
                 num_heads,
                 qkv_bias=True,
                 qk_scale=None,
                 attn_drop=0.,
                 proj_drop=0.):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        # define a parameter table of relative position bias
        # 2*Wh-1 * 2*Ww-1, nH
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1),
                        num_heads))

        # get pair-wise relative position index
        # for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = \
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        # Wh*Ww, Wh*Ww, 2
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()
        # shift to start from 0
        relative_coords[:, :, 0] += self.window_size[0] - 1
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.register_buffer('relative_position_index',
                             relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)

        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=.02)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of
                (num_windows, Wh*Ww, Wh*Ww) or None
        """
        B_, N, C = x.shape
        qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads,
                                  C // self.num_heads).permute(2, 0, 3, 1, 4)
        # make torchscript happy (cannot use tensor as tuple)
        q, k, v = qkv[0], qkv[1], qkv[2]

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        relative_position_bias = self.relative_position_bias_table[
            self.relative_position_index.view(-1)].view(
                self.window_size[0] * self.window_size[1],
                self.window_size[0] * self.window_size[1],
                -1)  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(
            2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nW = mask.shape[0]
            attn = attn.view(B_ // nW, nW, self.num_heads, N,
                             N) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, N, N)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, window_size={self.window_size}, ' \
               f'num_heads={self.num_heads}'


class SwinTransformerBlock(nn.Module):
    r""" Swin Transformer Block.
    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        num_heads (int): Number of attention heads.
        window_size (int): Window size.
        shift_size (int): Shift size for SW-MSA.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias
            to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk
            scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        norm_layer (nn.Module, optional): Normalization layer.
            Default: nn.LayerNorm
    """

    def __init__(self,
                 dim,
                 input_resolution,
                 num_heads,
                 window_size=7,
                 shift_size=0,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 act_layer=nn.GELU,
                 norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution,
            # we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)
        assert 0 <= self.shift_size < self.window_size, \
            'shift_size must in 0-window_size'

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim,
            window_size=to_2tuple(self.window_size),
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop)

        self.drop_path = DropPath(
            drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop)

        if self.shift_size > 0:
            attn_mask = self.calculate_mask(self.input_resolution)
        else:
            attn_mask = None

        self.register_buffer('attn_mask', attn_mask)

    def calculate_mask(self, x_size):
        # calculate attention mask for SW-MSA
        """Calculate attention mask for SW-MSA.

        Args:
            x_size (tuple[int]): Resolution of input feature.

        Returns:
            Tensor: Attention mask
        """
        H, W = x_size
        img_mask = torch.zeros((1, H, W, 1))  # 1 H W 1
        h_slices = (slice(0, -self.window_size),
                    slice(-self.window_size,
                          -self.shift_size), slice(-self.shift_size, None))
        w_slices = (slice(0, -self.window_size),
                    slice(-self.window_size,
                          -self.shift_size), slice(-self.shift_size, None))
        cnt = 0
        for h in h_slices:
            for w in w_slices:
                img_mask[:, h, w, :] = cnt
                cnt += 1

        mask_windows = window_partition(
            img_mask, self.window_size)  # nW, window_size, window_size, 1
        mask_windows = mask_windows.view(-1,
                                         self.window_size * self.window_size)
        attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
        attn_mask = attn_mask.masked_fill(attn_mask != 0,
                                          float(-100.0)).masked_fill(
                                              attn_mask == 0, float(0.0))

        return attn_mask

    def forward(self, x, x_size):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).
            x_size (tuple[int]): Resolution of input feature.

        Returns:
            Tensor: Forward results.
        """
        H, W = x_size
        B, L, C = x.shape
        # assert L == H * W, "input feature has wrong size"

        shortcut = x
        x = self.norm1(x)
        x = x.view(B, H, W, C)

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(
                x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
        else:
            shifted_x = x

        # partition windows
        x_windows = window_partition(
            shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size,
                                   C)  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA (to be compatible for testing on images
        # whose shapes are the multiple of window size
        if self.input_resolution == x_size:
            attn_windows = self.attn(
                x_windows,
                mask=self.attn_mask)  # nW*B, window_size*window_size, C
        else:
            attn_windows = self.attn(
                x_windows, mask=self.calculate_mask(x_size).to(x.device))

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size,
                                         self.window_size, C)
        shifted_x = window_reverse(attn_windows, self.window_size, H,
                                   W)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(
                shifted_x,
                shifts=(self.shift_size, self.shift_size),
                dims=(1, 2))
        else:
            x = shifted_x
        x = x.view(B, H * W, C)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, ' \
               f'input_resolution={self.input_resolution}, ' \
               f'num_heads={self.num_heads}, ' \
               f'window_size={self.window_size}, ' \
               f'shift_size={self.shift_size}, ' \
               f'mlp_ratio={self.mlp_ratio}'


class BasicLayer(nn.Module):
    """A basic Swin Transformer layer for one stage.

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias
            to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk
            scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate.
            Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer.
            Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the
            end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory.
            Default: False.
    """

    def __init__(self,
                 dim,
                 input_resolution,
                 depth,
                 num_heads,
                 window_size,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 norm_layer=nn.LayerNorm,
                 downsample=None,
                 use_checkpoint=False):

        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList([
            SwinTransformerBlock(
                dim=dim,
                input_resolution=input_resolution,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=0 if (i % 2 == 0) else window_size // 2,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i]
                if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer) for i in range(depth)
        ])

        # patch merging layer
        if downsample is not None:
            self.downsample = downsample(
                input_resolution, dim=dim, norm_layer=norm_layer)
        else:
            self.downsample = None

    def forward(self, x, x_size):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).
            x_size (tuple[int]): Resolution of input feature.

        Returns:
            Tensor: Forward results.
        """
        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x, x_size)
            else:
                x = blk(x, x_size)
        if self.downsample is not None:
            x = self.downsample(x)
        return x

    def extra_repr(self) -> str:
        return f'dim={self.dim}, ' \
               f'input_resolution={self.input_resolution}, ' \
               f'depth={self.depth}'


class RSTB(nn.Module):
    """Residual Swin Transformer Block (RSTB).

    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resolution.
        depth (int): Number of blocks.
        num_heads (int): Number of attention heads.
        window_size (int): Local window size.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            Default: 4.0
        qkv_bias (bool, optional): If True, add a learnable bias to
            query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale
            of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate.
            Default: 0.0
        norm_layer (nn.Module, optional): Normalization layer.
            Default: nn.LayerNorm
        downsample (nn.Module | None, optional): Downsample layer at the
            end of the layer. Default: None
        use_checkpoint (bool): Whether to use checkpointing to save memory.
            Default: False.
        img_size (int): Input image size. Default: 224
        patch_size (int): Patch size. Default: 4
        resi_connection (string): The convolutional block before
            residual connection. Default: '1conv'
    """

    def __init__(self,
                 dim,
                 input_resolution,
                 depth,
                 num_heads,
                 window_size,
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 norm_layer=nn.LayerNorm,
                 downsample=None,
                 use_checkpoint=False,
                 img_size=224,
                 patch_size=4,
                 resi_connection='1conv'):
        super(RSTB, self).__init__()

        self.dim = dim
        self.input_resolution = input_resolution

        self.residual_group = BasicLayer(
            dim=dim,
            input_resolution=input_resolution,
            depth=depth,
            num_heads=num_heads,
            window_size=window_size,
            mlp_ratio=mlp_ratio,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            drop=drop,
            attn_drop=attn_drop,
            drop_path=drop_path,
            norm_layer=norm_layer,
            downsample=downsample,
            use_checkpoint=use_checkpoint)

        if resi_connection == '1conv':
            self.conv = nn.Conv2d(dim, dim, 3, 1, 1)
        elif resi_connection == '3conv':
            # to save parameters and memory
            self.conv = nn.Sequential(
                nn.Conv2d(dim, dim // 4, 3, 1, 1),
                nn.LeakyReLU(negative_slope=0.2, inplace=True),
                nn.Conv2d(dim // 4, dim // 4, 1, 1, 0),
                nn.LeakyReLU(negative_slope=0.2, inplace=True),
                nn.Conv2d(dim // 4, dim, 3, 1, 1))

        self.patch_embed = PatchEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=0,
            embed_dim=dim,
            norm_layer=None)

        self.patch_unembed = PatchUnEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=0,
            embed_dim=dim,
            norm_layer=None)

    def forward(self, x, x_size):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (B, L, C).
            x_size (tuple[int]): Resolution of input feature.

        Returns:
            Tensor: Forward results.
        """
        return self.patch_embed(
            self.conv(
                self.patch_unembed(self.residual_group(x, x_size),
                                   x_size))) + x



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
from typing import Callable, List, Optional, Union

from mmengine.dataset import BaseDataset
from mmengine.fileio import get_file_backend, list_from_file

from ..registry import DATASETS


@DATASETS.register_module()
class BasicFramesDataset(BaseDataset):
    """BasicFramesDataset for open source projects in OpenMMLab/MMagic.

    This dataset is designed for low-level vision tasks with frames,
    such as video super-resolution and video frame interpolation.

    The annotation file is optional.

    If use annotation file, the annotation format can be shown as follows.

    .. code-block:: none

        Case 1 (Vid4):

            calendar 41
            city 34
            foliage 49
            walk 47

        Case 2 (REDS):

            000/00000000.png (720, 1280, 3)
            000/00000001.png (720, 1280, 3)

        Case 3 (Vimeo90k):

            00001/0266 (256, 448, 3)
            00001/0268 (256, 448, 3)

    Args:
        ann_file (str): Annotation file path. Defaults to ''.
        metainfo (dict, optional): Meta information for dataset, such as class
            information. Defaults to None.
        data_root (str, optional): The root directory for ``data_prefix`` and
            ``ann_file``. Defaults to None.
        data_prefix (dict, optional): Prefix for training data. Defaults to
            dict(img='', gt='').
        pipeline (list, optional): Processing pipeline. Defaults to [].
        test_mode (bool, optional): ``test_mode=True`` means in test phase.
            Defaults to False.
        filename_tmpl (str): Template for each filename. Note that the
            template excludes the file extension. Default: '{}'.
        search_key (str): The key used for searching the folder to get
            data_list. Default: 'gt'.
        backend_args (dict, optional): Arguments to instantiate the prefix of
            uri corresponding backend. Defaults to None.
        depth (int): The depth of path. Default: 1
        num_input_frames (None | int): Number of input frames. Default: None.
        num_output_frames (None | int): Number of output frames. Default: None.
        fixed_seq_len (None | int): The fixed sequence length.
            If None, BasicFramesDataset will obtain the length of each
            sequence.
            Default: None.
        load_frames_list (dict): Load frames list for each key.
            Default: dict().

    Examples:

        Assume the file structure as the following:

        mmagic (root)
        ├── mmagic
        ├── tools
        ├── configs
        ├── data
        │   ├── Vid4
        │   │   ├── BIx4
        │   │   │   ├── city
        │   │   │   │   ├── img1.png
        │   │   ├── GT
        │   │   │   ├── city
        │   │   │   │   ├── img1.png
        │   │   ├── meta_info_Vid4_GT.txt
        │   ├── places
        │   │   ├── sequences
        |   |   |   ├── 00001
        │   │   │   │   ├── 0389
        │   │   │   │   │   ├── img1.png
        │   │   │   │   │   ├── img2.png
        │   │   │   │   │   ├── img3.png
        │   │   ├── tri_trainlist.txt

        Case 1: Loading Vid4 dataset for training a VSR model.

        .. code-block:: python

            dataset = BasicFramesDataset(
                ann_file='meta_info_Vid4_GT.txt',
                metainfo=dict(dataset_type='vid4', task_name='vsr'),
                data_root='data/Vid4',
                data_prefix=dict(img='BIx4', gt='GT'),
                pipeline=[],
                depth=2,
                num_input_frames=5)

        Case 2: Loading Vimeo90k dataset for training a VFI model.

        .. code-block:: python

            dataset = BasicFramesDataset(
                ann_file='tri_trainlist.txt',
                metainfo=dict(dataset_type='vimeo90k', task_name='vfi'),
                data_root='data/vimeo-triplet',
                data_prefix=dict(img='sequences', gt='sequences'),
                pipeline=[],
                depth=2,
                load_frames_list=dict(
                    img=['img1.png', 'img3.png'], gt=['img2.png']))

        See more details in unittest
            tests/test_datasets/test_base_frames_dataset.py
                TestFramesDatasets().test_version_1_method()
    """

    METAINFO = dict(dataset_type='base_edit_dataset', task_name='editing')

    def __init__(self,
                 ann_file: str = '',
                 metainfo: Optional[dict] = None,
                 data_root: Optional[str] = None,
                 data_prefix: dict = dict(img=''),
                 pipeline: List[Union[dict, Callable]] = [],
                 test_mode: bool = False,
                 filename_tmpl: dict = dict(),
                 search_key: Optional[str] = None,
                 backend_args: Optional[dict] = None,
                 depth: int = 1,
                 num_input_frames: Optional[int] = None,
                 num_output_frames: Optional[int] = None,
                 fixed_seq_len: Optional[int] = None,
                 load_frames_list: dict = dict(),
                 **kwargs):

        for key in data_prefix:
            if key not in filename_tmpl:
                filename_tmpl[key] = '{}'

        if search_key is None:
            keys = list(data_prefix.keys())
            search_key = keys[0]
        self.search_key = search_key
        self.filename_tmpl = filename_tmpl
        self.use_ann_file = (ann_file != '')
        if backend_args is None:
            self.backend_args = None
        else:
            self.backend_args = backend_args.copy()
        self.depth = depth
        self.seq_lens = dict(fixed_seq_len=fixed_seq_len)
        self.num_input_frames = num_input_frames
        self.num_output_frames = num_output_frames
        self.load_frames_list = load_frames_list
        self.file_backend = get_file_backend(
            uri=data_root, backend_args=backend_args)

        super().__init__(
            ann_file=ann_file,
            metainfo=metainfo,
            data_root=data_root,
            data_prefix=data_prefix,
            pipeline=pipeline,
            test_mode=test_mode,
            **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load data list from folder or annotation file.

        Returns:
            list[dict]: A list of annotation.
        """

        path_list = self._get_path_list()
        self._set_seq_lens()

        data_list = []
        for path in path_list:
            basename, _ = osp.splitext(path)
            sequence_length = self.seq_lens['fixed_seq_len']
            if sequence_length is None:
                sequence_length = self.seq_lens[path.split(os.sep)[0]]
            data = dict(
                key=basename,
                num_input_frames=self.num_input_frames,
                num_output_frames=self.num_output_frames,
                sequence_length=sequence_length)
            for key in self.data_prefix:
                if key in self.load_frames_list:
                    folder = osp.join(self.data_prefix[key], path)
                    data[f'{key}_path'] = self._get_frames_list(key, folder)
                    # The list of frames has been loaded,
                    # ``sequence_length`` is useless
                    # Avoid loading frames by ``sequence_length`` in pipeline
                    data['sequence_length'] = None
                    # overwrite ``num_input_frames`` and ``num_output_frames``
                    if key == 'img':
                        data['num_input_frames'] = len(data[f'{key}_path'])
                    elif key == 'gt':
                        data['num_output_frames'] = len(data[f'{key}_path'])
                else:
                    data[f'{key}_path'] = self.data_prefix[key]
            data_list.append(data)

        return data_list

    def _get_path_list(self):
        """Get list of paths from annotation file or folder of dataset.

        Returns:
            list[str]: A list of paths.
        """

        if self.use_ann_file:
            path_list = self._get_path_list_from_ann()
        else:
            path_list = self._get_path_list_from_folder(depth=self.depth)

        return path_list

    def _get_path_list_from_ann(self):
        """Get list of paths from annotation file.

        Returns:
            list[str]: A list of paths.
        """

        ann_list = list_from_file(
            self.ann_file, backend_args=self.backend_args)
        path_list = []
        for ann in ann_list:
            if ann.isspace() or ann == '':
                continue
            path = ann.split(' ')[0]
            # Compatible with Windows file systems
            path = path.replace('/', os.sep)
            splitted_path = path.split(os.sep)
            if self.seq_lens['fixed_seq_len'] is None:
                self.seq_lens[splitted_path[0]] = 0
            ann_depth = len(splitted_path)
            if self.depth > ann_depth:
                # desire "folder/file", but the ann_file provides "folder".
                sub_path_list = self._get_path_list_from_folder(
                    sub_folder=path,
                    need_ext=False,
                    depth=self.depth - ann_depth)
                path_list.extend(
                    [path + os.sep + sub_path for sub_path in sub_path_list])
            elif self.depth < ann_depth:
                # desire "folder", while the ann_file provides "folder/file".
                desire_path = f'{os.sep}'.join(splitted_path[:self.depth])
                if desire_path not in path_list:
                    path_list.append(desire_path)
            else:
                # desire "folder/file" and the ann_file provides "folder/file".
                # or desire "folder" and the ann_file provides "folder".
                path_list.append(path)

        return path_list

    def _get_path_list_from_folder(self,
                                   sub_folder=None,
                                   need_ext=True,
                                   depth=1):
        """Get list of paths from folder.

        Args:
            sub_folder (None | str): The path of sub_folder. Default: None.
            need_ext (bool): Whether need ext. Default: True.
            depth (int): Residual depth of path, recursively called to
                ``depth == 1``. Default: 1

        Returns:
            list[str]: A list of paths.
        """

        folder = self.data_prefix[self.search_key]
        tmpl = self.filename_tmpl[self.search_key].format('')
        path_list = []
        if sub_folder:
            folder = osp.join(folder, sub_folder)
        listdir = list(self.file_backend.list_dir_or_file(dir_path=folder))
        listdir.sort()
        for path in listdir:
            basename, ext = osp.splitext(path)
            if not (sub_folder or self.seq_lens['fixed_seq_len']):
                self.seq_lens[basename] = 0
            if depth > 1:
                sub_path_list = self._get_path_list_from_folder(
                    sub_folder=path)
                path_list.extend(
                    [path + os.sep + sub_path for sub_path in sub_path_list])
            elif basename.endswith(tmpl):
                if need_ext:
                    path = path.replace(tmpl + ext, ext)
                else:
                    path = path.replace(tmpl + ext, '')
                path_list.append(path)

        return path_list

    def _set_seq_lens(self):
        """Get sequence lengths."""

        if self.seq_lens['fixed_seq_len']:
            return
        folder = self.data_prefix[self.search_key]
        for key in self.seq_lens.keys():
            if key == 'fixed_seq_len':
                continue
            path = osp.join(folder, key)
            num_frames = len(list(self.file_backend.list_dir_or_file(path)))
            self.seq_lens[key] = num_frames

    def _get_frames_list(self, key, folder):
        """Obtain list of frames.

        Args:
            key (str): The key of frames list, e.g. ``img``, ``gt``.
            folder (str): Folder of frames.

        Return:
            list[str]: The paths list of frames.
        """

        if 'all' in self.load_frames_list[key]:
            # load all
            files = list(self.file_backend.list_dir_or_file(dir_path=folder))
        else:
            files = self.load_frames_list[key]

        files.sort()
        tmpl = self.filename_tmpl[key]
        files = [tmpl.format(file) for file in files]
        paths = [osp.join(folder, file) for file in files]

        return paths



# Copyright (c) OpenMMLab. All rights reserved.
import mmcv
import numpy as np
from mmcv.transforms import BaseTransform

from mmagic.registry import TRANSFORMS


@TRANSFORMS.register_module()
class Normalize(BaseTransform):
    """Normalize images with the given mean and std value.

    Required keys are the keys in attribute "keys", added or modified keys are
    the keys in attribute "keys" and these keys with postfix '_norm_cfg'.
    It also supports normalizing a list of images.

    Args:
        keys (Sequence[str]): The images to be normalized.
        mean (np.ndarray): Mean values of different channels.
        std (np.ndarray): Std values of different channels.
        to_rgb (bool): Whether to convert channels from BGR to RGB.
            Default: False.
        save_original (bool): Whether to save original images. Default: False.
    """

    def __init__(self, keys, mean, std, to_rgb=False, save_original=False):
        self.keys = keys
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)
        self.to_rgb = to_rgb
        self.save_original = save_original

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        for key in self.keys:
            if isinstance(results[key], list):
                if self.save_original:
                    results[key + '_unnormalised'] = [
                        v.copy() for v in results[key]
                    ]
                results[key] = [
                    mmcv.imnormalize(v, self.mean, self.std, self.to_rgb)
                    for v in results[key]
                ]
            else:
                if self.save_original:
                    results[key + '_unnormalised'] = results[key].copy()
                results[key] = mmcv.imnormalize(results[key], self.mean,
                                                self.std, self.to_rgb)

        results['img_norm_cfg'] = dict(
            mean=self.mean, std=self.std, to_rgb=self.to_rgb)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (f'(keys={self.keys}, mean={self.mean}, std={self.std}, '
                     f'to_rgb={self.to_rgb})')

        return repr_str


@TRANSFORMS.register_module()
class RescaleToZeroOne(BaseTransform):
    """Transform the images into a range between 0 and 1.

    Required keys are the keys in attribute "keys", added or modified keys are
    the keys in attribute "keys".
    It also supports rescaling a list of images.

    Args:
        keys (Sequence[str]): The images to be transformed.
    """

    def __init__(self, keys):
        self.keys = keys

    def transform(self, results):
        """transform function.

        Args:
            results (dict): A dict containing the necessary information and
                data for augmentation.

        Returns:
            dict: A dict containing the processed data and information.
        """
        for key in self.keys:
            if isinstance(results[key], list):
                results[key] = [
                    v.astype(np.float32) / 255. for v in results[key]
                ]
            else:
                results[key] = results[key].astype(np.float32) / 255.
        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(keys={self.keys})'



# Copyright (c) OpenMMLab. All rights reserved.
import os
from typing import Dict, List

import numpy as np
from mmengine import mkdir_or_exist
from PIL.Image import Image
from torchvision.utils import save_image

from .base_mmagic_inferencer import BaseMMagicInferencer, InputsType, PredType


class Text2ImageInferencer(BaseMMagicInferencer):
    """inferencer that predicts with text2image models."""

    func_kwargs = dict(
        preprocess=['text'],
        forward=[],
        visualize=['result_out_dir'],
        postprocess=[])

    extra_parameters = dict(height=None, width=None, seed=1)

    def preprocess(self, text: InputsType) -> Dict:
        """Process the inputs into a model-feedable format.

        Args:
            text(InputsType): text input for text-to-image model.

        Returns:
            result(Dict): Results of preprocess.
        """
        result = self.extra_parameters
        if type(text) is dict:
            result['text_prompts'] = text
        else:
            result['prompt'] = text

        return result

    def forward(self, inputs: InputsType) -> PredType:
        """Forward the inputs to the model."""
        image = self.model.infer(**inputs)['samples']

        return image

    def visualize(self,
                  preds: PredType,
                  result_out_dir: str = None) -> List[np.ndarray]:
        """Visualize predictions.

        Args:
            preds (List[Union[str, np.ndarray]]): Forward results
                by the inferencer.
            result_out_dir (str): Output directory of image.
                Defaults to ''.

        Returns:
            List[np.ndarray]: Result of visualize
        """
        if result_out_dir:
            mkdir_or_exist(os.path.dirname(result_out_dir))
            if type(preds) is list:
                preds = preds[0]
            if type(preds) is Image:
                preds.save(result_out_dir)
            else:
                save_image(preds, result_out_dir, normalize=True)

        return preds



_base_ = [
    '../_base_/datasets/comp1k.py', '../_base_/matting_default_runtime.py'
]

experiment_name = 'indexnet_mobv2_1xb16-78k_comp1k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

# model settings
model = dict(
    type='IndexNet',
    data_preprocessor=dict(
        type='MattorPreprocessor',
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        proc_trimap='rescale_to_zero_one',
    ),
    backbone=dict(
        type='SimpleEncoderDecoder',
        encoder=dict(
            type='IndexNetEncoder',
            in_channels=4,
            freeze_bn=True,
            init_cfg=dict(
                type='Pretrained',
                checkpoint='open-mmlab://mmedit/mobilenet_v2')),
        decoder=dict(type='IndexNetDecoder')),
    loss_alpha=dict(type='CharbonnierLoss', loss_weight=0.5, sample_wise=True),
    loss_comp=dict(
        type='CharbonnierCompLoss', loss_weight=1.5, sample_wise=True),
    test_cfg=dict(
        resize_method='interp',
        resize_mode='bicubic',
        size_divisor=32,
    ),
)

train_pipeline = [
    dict(type='LoadImageFromFile', key='alpha', color_type='grayscale'),
    dict(type='LoadImageFromFile', key='fg'),
    dict(type='LoadImageFromFile', key='bg'),
    dict(type='LoadImageFromFile', key='merged'),
    dict(type='GenerateTrimapWithDistTransform', dist_thr=20),
    dict(
        type='CropAroundUnknown',
        keys=['alpha', 'merged', 'fg', 'bg', 'trimap'],
        crop_sizes=[320, 480, 640],
        interpolations=['bicubic', 'bicubic', 'bicubic', 'bicubic',
                        'nearest']),
    dict(
        type='Resize',
        keys=['trimap'],
        scale=(320, 320),
        keep_ratio=False,
        interpolation='nearest'),
    dict(
        type='Resize',
        keys=['alpha', 'merged', 'fg', 'bg'],
        scale=(320, 320),
        keep_ratio=False,
        interpolation='bicubic'),
    dict(type='Flip', keys=['alpha', 'merged', 'fg', 'bg', 'trimap']),
    dict(type='PackInputs'),
]

test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='alpha',
        color_type='grayscale',
        save_original_img=True),
    dict(
        type='LoadImageFromFile',
        key='trimap',
        color_type='grayscale',
        save_original_img=True),
    dict(type='LoadImageFromFile', key='merged'),
    dict(type='PackInputs'),
]

train_dataloader = dict(
    batch_size=16,
    num_workers=8,
    dataset=dict(pipeline=train_pipeline),
)

val_dataloader = dict(
    batch_size=1,
    dataset=dict(pipeline=test_pipeline),
)

test_dataloader = val_dataloader

train_cfg = dict(
    type='IterBasedTrainLoop',
    max_iters=78000,
    val_interval=2600,
)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=1e-2),
    paramwise_cfg=dict(custom_keys={'encoder.layers': dict(lr_mult=0.01)}),
)
# learning policy
param_scheduler = dict(
    type='MultiStepLR',
    milestones=[52000, 67600],
    gamma=0.1,
    by_epoch=False,
)

# checkpoint saving
default_hooks = dict(checkpoint=dict(interval=2600, out_dir=save_dir))

# runtime settings
# inheritate from _base_



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from mmengine.model import BaseModule

from mmagic.models.archs import MultiLayerDiscriminator
from mmagic.registry import MODELS


@MODELS.register_module()
class GLDiscs(BaseModule):
    """Discriminators in Global&Local.

    This discriminator contains a local discriminator and a global
    discriminator as described in the original paper:
    Globally and locally Consistent Image Completion

    Args:
        global_disc_cfg (dict): Config dict to build global discriminator.
        local_disc_cfg (dict): Config dict to build local discriminator.
    """

    def __init__(self, global_disc_cfg, local_disc_cfg):
        super().__init__()
        self.global_disc = MultiLayerDiscriminator(**global_disc_cfg)
        self.local_disc = MultiLayerDiscriminator(**local_disc_cfg)

        self.fc = nn.Linear(2048, 1, bias=True)

    def forward(self, x):
        """Forward function.

        Args:
            x (tuple[torch.Tensor]): Contains global image and the local image
                patch.

        Returns:
            tuple[torch.Tensor]: Contains the prediction from discriminators \
                in global image and local image patch.
        """
        g_img, l_img = x
        g_pred = self.global_disc(g_img)
        l_pred = self.local_disc(l_img)

        pred = self.fc(torch.cat([g_pred, l_pred], dim=1))

        return pred

    def init_weights(self):
        """Init weights for models."""

        for m in self.modules():
            # Here, we only initialize the module with fc layer since the
            # conv and norm layers has been initialized in `ConvModule`.
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight.data, 0.0, 0.02)
                nn.init.constant_(m.bias.data, 0.0)

        self._is_init = True



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from mmcv.cnn import ConvModule


class UnetSkipConnectionBlock(nn.Module):
    """Construct a Unet submodule with skip connections, with the following.

    structure: downsampling - `submodule` - upsampling.

    Args:
        outer_channels (int): Number of channels at the outer conv layer.
        inner_channels (int): Number of channels at the inner conv layer.
        in_channels (int): Number of channels in input images/features. If is
            None, equals to `outer_channels`. Default: None.
        submodule (UnetSkipConnectionBlock): Previously constructed submodule.
            Default: None.
        is_outermost (bool): Whether this module is the outermost module.
            Default: False.
        is_innermost (bool): Whether this module is the innermost module.
            Default: False.
        norm_cfg (dict): Config dict to build norm layer. Default:
            `dict(type='BN')`.
        use_dropout (bool): Whether to use dropout layers. Default: False.
    """

    def __init__(self,
                 outer_channels,
                 inner_channels,
                 in_channels=None,
                 submodule=None,
                 is_outermost=False,
                 is_innermost=False,
                 norm_cfg=dict(type='BN'),
                 use_dropout=False):
        super().__init__()
        # cannot be both outermost and innermost
        assert not (is_outermost and is_innermost), (
            "'is_outermost' and 'is_innermost' cannot be True"
            'at the same time.')
        self.is_outermost = is_outermost
        assert isinstance(norm_cfg, dict), ("'norm_cfg' should be dict, but"
                                            f'got {type(norm_cfg)}')
        assert 'type' in norm_cfg, "'norm_cfg' must have key 'type'"
        # We use norm layers in the unet skip connection block.
        # Only for IN, use bias since it does not have affine parameters.
        use_bias = norm_cfg['type'] == 'IN'

        kernel_size = 4
        stride = 2
        padding = 1
        if in_channels is None:
            in_channels = outer_channels
        down_conv_cfg = dict(type='Conv2d')
        down_norm_cfg = norm_cfg
        down_act_cfg = dict(type='LeakyReLU', negative_slope=0.2)
        up_conv_cfg = dict(type='Deconv')
        up_norm_cfg = norm_cfg
        up_act_cfg = dict(type='ReLU')
        up_in_channels = inner_channels * 2
        up_bias = use_bias
        middle = [submodule]
        upper = []

        if is_outermost:
            down_act_cfg = None
            down_norm_cfg = None
            up_bias = True
            up_norm_cfg = None
            upper = [nn.Tanh()]
        elif is_innermost:
            down_norm_cfg = None
            up_in_channels = inner_channels
            middle = []
        else:
            upper = [nn.Dropout(0.5)] if use_dropout else []

        down = [
            ConvModule(
                in_channels=in_channels,
                out_channels=inner_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                bias=use_bias,
                conv_cfg=down_conv_cfg,
                norm_cfg=down_norm_cfg,
                act_cfg=down_act_cfg,
                order=('act', 'conv', 'norm'))
        ]
        up = [
            ConvModule(
                in_channels=up_in_channels,
                out_channels=outer_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                bias=up_bias,
                conv_cfg=up_conv_cfg,
                norm_cfg=up_norm_cfg,
                act_cfg=up_act_cfg,
                order=('act', 'conv', 'norm'))
        ]

        model = down + middle + up + upper

        self.model = nn.Sequential(*model)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """
        if self.is_outermost:
            return self.model(x)

        # add skip connections
        return torch.cat([x, self.model(x)], 1)



default_scope = 'mmagic'

randomness = dict(seed=2022, diff_rank_seed=True)
# env settings
dist_params = dict(backend='nccl')
# disable opencv multithreading to avoid system being overloaded
opencv_num_threads = 0
# set multi-process start method as `fork` to speed up the training
mp_start_method = 'fork'

# configure for default hooks
default_hooks = dict(
    # record time of every iteration.
    timer=dict(type='IterTimerHook'),
    # print log every 100 iterations.
    logger=dict(type='LoggerHook', interval=100, log_metric_by_epoch=False),
    # save checkpoint per 10000 iterations
    checkpoint=dict(
        type='CheckpointHook',
        interval=10000,
        by_epoch=False,
        max_keep_ckpts=20,
        less_keys=['FID-Full-50k/fid', 'FID-50k/fid', 'swd/avg'],
        greater_keys=['IS-50k/is', 'ms-ssim/avg'],
        save_optimizer=True))

# config for environment
env_cfg = dict(
    # whether to enable cudnn benchmark.
    cudnn_benchmark=True,
    # set multi process parameters.
    mp_cfg=dict(mp_start_method='fork', opencv_num_threads=0),
    # set distributed parameters.
    dist_cfg=dict(backend='nccl'))

# set log level
log_level = 'INFO'
log_processor = dict(type='LogProcessor', by_epoch=False)

# load from which checkpoint
load_from = None

# whether to resume training from the loaded checkpoint
resume = None

# config for model wrapper
model_wrapper_cfg = dict(
    type='MMSeparateDistributedDataParallel',
    broadcast_buffers=False,
    find_unused_parameters=False)

# set visualizer
vis_backends = [dict(type='VisBackend')]
visualizer = dict(type='Visualizer', vis_backends=vis_backends)

# config for training
train_cfg = dict(by_epoch=False, val_begin=1, val_interval=10000)

# config for val
val_cfg = dict(type='MultiValLoop')
val_evaluator = dict(type='Evaluator')

# config for test
test_cfg = dict(type='MultiTestLoop')
test_evaluator = dict(type='Evaluator')

# config for optim_wrapper_constructor
optim_wrapper = dict(constructor='MultiOptimWrapperConstructor')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch

from mmagic.utils import ConfigType
from .colorization_inferencer import ColorizationInferencer
from .conditional_inferencer import ConditionalInferencer
from .controlnet_animation_inferencer import ControlnetAnimationInferencer
from .eg3d_inferencer import EG3DInferencer
from .image_super_resolution_inferencer import ImageSuperResolutionInferencer
from .inpainting_inferencer import InpaintingInferencer
from .matting_inferencer import MattingInferencer
from .text2image_inferencer import Text2ImageInferencer
from .translation_inferencer import TranslationInferencer
from .unconditional_inferencer import UnconditionalInferencer
from .video_interpolation_inferencer import VideoInterpolationInferencer
from .video_restoration_inferencer import VideoRestorationInferencer

__all__ = [
    'ColorizationInferencer', 'ConditionalInferencer', 'EG3DInferencer',
    'InpaintingInferencer', 'MattingInferencer',
    'ImageSuperResolutionInferencer', 'Text2ImageInferencer',
    'TranslationInferencer', 'UnconditionalInferencer',
    'VideoInterpolationInferencer', 'VideoRestorationInferencer',
    'ControlnetAnimationInferencer'
]


class Inferencers:
    """Class to assign task to different inferencers.

    Args:
        task (str): Inferencer task.
        config (str or ConfigType): Model config or the path to it.
        ckpt (str, optional): Path to the checkpoint.
        device (str, optional): Device to run inference. If None, the best
            device will be automatically used.
        seed (int): The random seed used in inference. Defaults to 2022.
    """

    def __init__(self,
                 task: Optional[str] = None,
                 config: Optional[Union[ConfigType, str]] = None,
                 ckpt: Optional[str] = None,
                 device: torch.device = None,
                 extra_parameters: Optional[Dict] = None,
                 seed: int = 2022) -> None:
        self.task = task
        if self.task in ['conditional', 'Conditional GANs']:
            self.inferencer = ConditionalInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['colorization', 'Colorization']:
            self.inferencer = ColorizationInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['unconditional', 'Unconditional GANs']:
            self.inferencer = UnconditionalInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['matting', 'Matting']:
            self.inferencer = MattingInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['inpainting', 'Inpainting']:
            self.inferencer = InpaintingInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['translation', 'Image2Image']:
            self.inferencer = TranslationInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['Image super-resolution', 'Image Super-Resolution']:
            self.inferencer = ImageSuperResolutionInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['video_restoration', 'Video Super-Resolution']:
            self.inferencer = VideoRestorationInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['video_interpolation', 'Video Interpolation']:
            self.inferencer = VideoInterpolationInferencer(
                config, ckpt, device, extra_parameters)
        elif self.task in [
                'text2image', 'Text2Image', 'Text2Image, Image2Image'
        ]:
            self.inferencer = Text2ImageInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['3D_aware_generation', '3D-aware Generation']:
            self.inferencer = EG3DInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        elif self.task in ['controlnet_animation']:
            self.inferencer = ControlnetAnimationInferencer(config)
        elif self.task in [
                'Image Restoration', 'Denoising, Deblurring, Deraining',
                'Image Super-Resolution, Image denoising, JPEG compression '
                'artifact reduction'
        ]:
            self.inferencer = ImageSuperResolutionInferencer(
                config, ckpt, device, extra_parameters, seed=seed)
        else:
            raise ValueError(f'Unknown inferencer task: {self.task}')

    def __call__(self, **kwargs) -> Union[Dict, List[Dict]]:
        """Call the inferencer.

        Args:
            kwargs: Keyword arguments for the inferencer.

        Returns:
            Union[Dict, List[Dict]]: Results of inference pipeline.
        """
        return self.inferencer(**kwargs)

    def get_extra_parameters(self) -> List[str]:
        """Each inferencer may has its own parameters. Call this function to
        get these parameters.

        Returns:
            List[str]: List of unique parameters.
        """
        return self.inferencer.get_extra_parameters()



# Copyright (c) OpenMMLab. All rights reserved.
import re
from typing import Any, List, Optional, Union

import torch.nn as nn
from mmengine import print_log
from torch import Tensor


class LoRALinear(nn.Module):
    """Linear layer for LoRA.

    Args:
        in_feat (int): Number of input features.
        out_feat (int): Number of output features.
        rank (int): The rank of LoRA.
    """

    def __init__(self, in_feat: int, out_feat: int, rank: int = 4):
        super().__init__()
        self.rank = rank
        assert rank < min(in_feat, out_feat)

        self.down = nn.Linear(in_feat, rank, bias=False)
        self.up = nn.Linear(rank, out_feat, bias=False)

        nn.init.normal_(self.down.weight, std=1 / rank)
        nn.init.zeros_(self.up.weight)

    def forward(self, x: Tensor) -> Tensor:
        ori_type = x.dtype
        dtype = self.down.weight.dtype

        out = self.down(x.to(dtype))
        out = self.up(out)
        return out.to(ori_type)


class LoRAWrapper(nn.Module):
    """Wrapper for LoRA layer.

    Args:
        module (nn.Module): The module to be wrapped.
        in_feat (int): Number of input features.
        out_feat (int): Number of output features.
        rank (int): The rank of LoRA.
        scale (float): The scale of LoRA feature.
        names (Union[str, List[str]], optional): The name of LoRA layers. If
            you want to add multi LoRA for one module, names for each LoRA
            mapping must be defined.
    """

    def __init__(self,
                 module: nn.Module,
                 in_feat: int,
                 out_feat: int,
                 rank: int,
                 scale: float = 1,
                 names: Optional[Union[str, List[str]]] = None):
        super().__init__()

        # NOTE: LoRA for linear layer, LoCON will coming soon~
        assert isinstance(
            module,
            nn.Linear), ('Only Support LoRA for linear layer currently. '
                         'LoCON will coming soon~')
        self.wrapped = module

        if names is not None:
            # set a list of LoRAs
            if not isinstance(names, list):
                names = [names]
            if isinstance(rank, list):
                assert len(rank) == len(names)
            else:
                rank = [rank] * len(names)
            if isinstance(scale, list):
                assert len(scale) == len(names)
            else:
                scale = [scale] * len(names)

            self.names = names
            self.lora_mapping = dict()
            self.scale = dict()
            self.enable = dict()
            self.rank = dict()
            for n, r, s in zip(names, rank, scale):
                self.lora_mapping[n] = LoRALinear(in_feat, out_feat, r)
                self.scale_dict[n] = s
                self.enable[n] = True
                self.rank[n] = r
            self.lora_mapping = nn.ModuleDict(self.lora_mapping)

        else:
            # set single LoRA
            self.names = None
            self.lora_mapping = LoRALinear(in_feat, out_feat, rank)
            self.scale = scale
            self.enable = True
            self.rank = rank

        self.in_feat, self.out_feat = in_feat, out_feat

    def add_lora(self,
                 name: str,
                 rank: int,
                 scale: float = 1,
                 state_dict: Optional[dict] = None):
        """Add LoRA mapping.

        Args:
            name (str): The name of added LoRA.
            rank (int): The rank of added LoRA.
            scale (float, optional): The scale of added LoRA. Defaults to 1.
            state_dict (dict, optional): The state dict of added LoRA.
                Defaults to None.
        """
        mapping_to_add = LoRALinear(self.in_feat, self.out_feat, rank)
        if state_dict is not None:
            mapping_to_add.load_state_dict(mapping_to_add)
        # move to device and type
        mapping_to_add.to(self.lora_mapping.weight.dtype)

        if isinstance(self.names, list):
            self.names.append(name)
            self.lora_mapping[name] = mapping_to_add
            self.scale[name] = scale
            self.enable[name] = True
            self.rank[name] = rank
        else:
            self.names = ['orig', name]
            self.lora_mapping = nn.ModuleDict({
                'orig': self.lora_mapping,
                name: mapping_to_add
            })
            self.scale = {'orig': self.scale, name: scale}
            self.enable = {'orig': self.enable, name: True}
            self.rank = {'orig': self.rank, name: rank}
            print_log(
                'The original LoRA mapping do not have name, '
                'save as \'orig\'.', 'current')
        print_log(f'Add LoRA \'{name}\' with rank {rank} and scale {scale}.',
                  'current')

    def _set_value(self,
                   attr_name: str,
                   value: Any,
                   name: Optional[str] = None):
        """Set value of attribute.

        Args:
            attr_name (str): The name of attribute to be set value.
            value (Any): The value to be set.
            name (str, optional): The name of field in `attr_name`. If
                passed, will set value to `attr_name[name]`. Defaults to None.
        """
        attr = getattr(self, attr_name)

        if isinstance(attr, dict):
            if name is None:
                attr = {k: value for k in self.names}
                print_log(f'Set all value in \'{attr_name}\' as \'{value}\'.',
                          'current')
            else:
                attr[name] = value
                print_log(f'Set \'{attr_name}[{name}]\' as \'{value}\'.',
                          'current')
        else:
            attr = value
            print_log(f'Set \'{attr_name}\' as \'{value}\'.', 'current')

        setattr(self, attr_name, attr)

    def set_scale(self, scale: float, name: Optional[str] = None):
        """Set LoRA scale.

        Args:
            scale (float): The scale to be set.
            name (str, optional): The name of LoRA to be set. Defaults to None.
        """
        self._set_value('scale', scale, name)

    def set_enable(self, name: Optional[str] = None):
        """Enable LoRA for the current layer.

        Args:
            name (str, optional): The name of LoRA to be set. Defaults to None.
        """
        self._set_value('enable', True, name)

    def set_disable(self, name: Optional[str] = None):
        """Disable LoRA for the current layer.

        Args:
            name (str, optional): The name of LoRA to be set. Defaults to None.
        """
        self._set_value('enable', False, name)

    def forward_lora_mapping(self, x: Tensor) -> Tensor:
        """Forward LoRA mapping.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        mapping_out = 0
        if isinstance(self.lora_mapping, dict):
            for name in self.names:
                scale = self.scale[name]
                mapping_layer = self.lora_mapping[name]
                enable = self.enable[name]

                if enable:
                    mapping_out = scale * mapping_layer(x)
        else:
            if self.enable:
                mapping_out = self.scale * self.lora_mapping(x)
        return mapping_out

    def forward(self, x: Tensor) -> Tensor:
        """Forward and add LoRA mapping.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        mapping_out = self.forward_lora_mapping(x)
        return mapping_out + self.wrapped(x)

    @classmethod
    def wrap_lora(cls, module, rank=4, scale=1, names=None, state_dict=None):
        """Wrap LoRA.

        Use case:
        >>> linear = nn.Linear(2, 4)
        >>> lora_linear = LoRAWrapper.wrap_lora(linear, 4, 1)

        Args:
            module (nn.Module): The module to add LoRA.
            rank (int): The rank for LoRA.
            scale (float):

        Returns:
            LoRAWrapper:
        """
        assert isinstance(module,
                          nn.Linear), 'Only support LoRA for Linear Layer'
        in_feat = module.weight.shape[1]
        out_feat = module.weight.shape[0]
        lora = LoRAWrapper(module, in_feat, out_feat, rank, scale, names)

        return lora


def replace_module(parent_module: nn.Module, child_name: str,
                   new_module: nn.Module):
    """Replace module in parent module."""
    setattr(parent_module, child_name, new_module)


def get_submodule(module: nn.Module, key: str):
    """Get submodule by key."""
    target_name = key.split('.')[-1]
    parent = module.get_submodule('.'.join(key.split('.')[:-1]))
    target = module.get_submodule(key)
    return parent, target, target_name


def set_lora(module: nn.Module,
             config: dict,
             verbose: bool = True) -> nn.Module:
    """Set LoRA for module.

    Use case:
    >>> 1. set all lora with same parameters
    >>> lora_config = dict(
    >>>     rank=4,
    >>>     scale=1,
    >>>     target_modules=['to_q', 'to_k', 'to_v'])

    >>> 2. set lora with different parameters
    >>> lora_config = dict(
    >>>     rank=4,
    >>>     scale=1,
    >>>     target_modules=[
    >>>         # set `to_q` the default parameters
    >>>         'to_q',
    >>>         # set `to_k` the defined parameters
    >>>         dict(target_module='to_k', rank=8, scale=1),
    >>>         # set `to_v` the defined `rank` and default `scale`
    >>>         dict(target_module='to_v', rank=16)
    >>>     ])

    Args:
        module (nn.Module): The module to set LoRA.
        config (dict): The config dict.
        verbose (bool): Whether to print log. Defaults to True.
    """
    default_rank = config.get('rank', 4)
    default_scale = config.get('scale', 1)
    target_modules = config['target_modules']
    if not isinstance(target_modules, list):
        target_modules = [target_modules]

    keys = [k for k, _ in module.named_modules()]

    for k in keys:
        for target_module in target_modules:
            if isinstance(target_module, str):
                module_name = target_module
                rank = default_rank
                scale = default_scale
                # pretrained_path = None

            elif isinstance(target_module, dict):
                module_name = target_module['target_module']
                rank = target_module.get('rank', default_rank)
                scale = target_module.get('scale', default_scale)
                # pretrained_path = target_module.get('pretrained_path', None)

            else:
                raise TypeError('Only support dict or string type for '
                                'target_modules')
            # match keys
            if re.fullmatch(module_name, k):
                if verbose:
                    print_log(
                        f'Set LoRA for \'{k}\' with '
                        f'regularization expression match \'{module_name}\'.',
                        'current')
            elif k.endswith(module_name):
                if verbose:
                    print_log(
                        f'Set LoRA for \'{k}\' with '
                        f'suffix match \'{module_name}\'.', 'current')
            else:
                continue

            parent, target, target_name = get_submodule(module, k)
            new_module = LoRAWrapper.wrap_lora(target, rank=rank, scale=scale)
            replace_module(parent, target_name, new_module)
    return module


def set_only_lora_trainable(module: nn.Module) -> nn.Module:
    """Set only LoRA modules trainable."""
    for n, m in module.named_children():
        if isinstance(m, LoRAWrapper):
            m.lora_mapping.requires_grad_(True)
        elif isinstance(m, nn.Module):
            m.requires_grad_(False)
            set_only_lora_trainable(m)

    return module


def set_lora_enable(module: nn.Module) -> nn.Module:
    """Enable LoRA modules."""
    for n, m in module.named_children():
        if isinstance(m, LoRAWrapper):
            m.set_enable()
        elif isinstance(m, nn.Module):
            set_lora_enable(m)
    return module


def set_lora_disable(module: nn.Module) -> nn.Module:
    """Disable LoRA modules."""
    for n, m in module.named_children():
        if isinstance(m, LoRAWrapper):
            m.set_disable()
        elif isinstance(m, nn.Module):
            set_lora_disable(m)
    return module



_base_ = ['../_base_/default_runtime.py']

experiment_name = 'swinir_psnr-x2s64w8d6e180_8xb4-lr1e-4-600k_df2k-ost'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

scale = 2
img_size = 64

# model settings
model = dict(
    type='BaseEditModel',
    generator=dict(
        type='SwinIRNet',
        upscale=scale,
        in_chans=3,
        img_size=img_size,
        window_size=8,
        img_range=1.0,
        depths=[6, 6, 6, 6, 6, 6],
        embed_dim=180,
        num_heads=[6, 6, 6, 6, 6, 6],
        mlp_ratio=2,
        upsampler='nearest+conv',
        resi_connection='1conv'),
    pixel_loss=dict(type='L1Loss', loss_weight=1.0, reduction='mean'),
    data_preprocessor=dict(
        type='DataPreprocessor', mean=[0., 0., 0.], std=[255., 255., 255.]))

test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackInputs')
]

test_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='realsrset', task_name='realsr'),
        data_root='data/RealSRSet+5images',
        data_prefix=dict(img='', gt=''),
        pipeline=test_pipeline))

test_evaluator = [dict(type='NIQE', input_order='CHW', convert_to='Y')]

test_cfg = dict(type='TestLoop')



# Copyright (c) OpenMMLab. All rights reserved.
import pickle
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.nn.functional as F
from mmengine import Config, MessageHub, print_log
from mmengine.model import is_model_wrapper
from mmengine.optim import OptimWrapper, OptimWrapperDict
from torch import Tensor

from mmagic.models.utils import get_module_device
from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils import ForwardInputs, SampleList
from ...base_models import BaseGAN
from ...utils import set_requires_grad

ModelType = Union[Dict, nn.Module]
TrainInput = Union[dict, Tensor]


@MODELS.register_module()
class SinGAN(BaseGAN):
    """SinGAN.

    This model implement the single image generative adversarial model proposed
    in: Singan: Learning a Generative Model from a Single Natural Image,
    ICCV'19.

    Notes for training:

    - This model should be trained with our dataset ``SinGANDataset``.
    - In training, the ``total_iters`` arguments is related to the number of
      scales in the image pyramid and ``iters_per_scale`` in the ``train_cfg``.
      You should set it carefully in the training config file.

    Notes for model architectures:

    - The generator and discriminator need ``num_scales`` in initialization.
      However, this arguments is generated by ``create_real_pyramid`` function
      from the ``singan_dataset.py``. The last element in the returned list
      (``stop_scale``) is the value for ``num_scales``. Pay attention that this
      scale is counted from zero. Please see our tutorial for SinGAN to obtain
      more details or our standard config for reference.

    Args:
        generator (ModelType): The config or model of the generator.
        discriminator (Optional[ModelType]): The config or model of the
            discriminator. Defaults to None.
        data_preprocessor (Optional[Union[dict, Config]]): The pre-process
            config or :class:`~mmagic.models.DataPreprocessor`.
        generator_steps (int): The number of times the generator is completely
            updated before the discriminator is updated. Defaults to 1.
        discriminator_steps (int): The number of times the discriminator is
            completely updated before the generator is updated. Defaults to 1.
        num_scales (int): The number of scales/stages in generator/
            discriminator. Note that this number is counted from zero, which
            is the same as the original paper. Defaults to None.
        iters_per_scale (int): The training iteration for each resolution
            scale. Defaults to 2000.
        noise_weight_init (float): The initialize weight of fixed noise.
            Defaults to 0.1
        lr_scheduler_args (Optional[dict]): Arguments for learning schedulers.
            Note that in SinGAN, we use MultiStepLR, which is the same as the
            original paper. If not passed, no learning schedule will be used.
            Defaults to None.
        test_pkl_data (Optional[str]): The path of pickle file which contains
            fixed noise and noise weight. This is must for test. Defaults to
            None.
        ema_config (Optional[Dict]): The config for generator's exponential
            moving average setting. Defaults to None.
    """

    def __init__(self,
                 generator: ModelType,
                 discriminator: Optional[ModelType] = None,
                 data_preprocessor: Optional[Union[dict, Config]] = None,
                 generator_steps: int = 1,
                 discriminator_steps: int = 1,
                 num_scales: Optional[int] = None,
                 iters_per_scale: int = 2000,
                 noise_weight_init: int = 0.1,
                 lr_scheduler_args: Optional[dict] = None,
                 test_pkl_data: Optional[str] = None,
                 ema_confg: Optional[dict] = None):
        super().__init__(generator, discriminator, data_preprocessor,
                         generator_steps, discriminator_steps, None, ema_confg)
        self.iters_per_scale = iters_per_scale
        self.noise_weight_init = noise_weight_init
        if lr_scheduler_args:
            self.lr_scheduler_args = deepcopy(lr_scheduler_args)
        else:
            self.lr_scheduler_args = None

        # related to train
        self.num_scales = num_scales
        self.curr_stage = -1
        self.noise_weights = [1]
        self.fixed_noises = []
        self.reals = []

        # related to test
        self.loaded_test_pkl = False
        self.pkl_data = test_pkl_data

    def load_test_pkl(self):
        """Load pickle for test."""
        if self.pkl_data is not None:
            with open(self.pkl_data, 'rb') as f:
                data = pickle.load(f)
                self.fixed_noises = self._from_numpy(data['fixed_noises'])
                self.noise_weights = self._from_numpy(data['noise_weights'])
                self.curr_stage = data['curr_stage']
            print_log(f'Load pkl data from {self.pkl_data}', 'current')
            self.pkl_data = self.pkl_data
            self.loaded_test_pkl = True

    def _from_numpy(self,
                    data: Tuple[list,
                                np.ndarray]) -> Tuple[Tensor, List[Tensor]]:
        """Convert input numpy array or list of numpy array to Tensor or list
        of Tensor.

        Args:
            data (Tuple[list, np.ndarray]): Input data to convert.

        Returns:
            Tuple[Tensor, List[Tensor]]: Converted Tensor or list of tensor.
        """
        if isinstance(data, list):
            return [self._from_numpy(x) for x in data]

        if isinstance(data, np.ndarray):
            data = torch.from_numpy(data)
            device = get_module_device(self.generator)
            data = data.to(device)
            return data

        return data

    def get_module(self, model: nn.Module, module_name: str) -> nn.Module:
        """Get an inner module from model.

        Since we will wrapper DDP for some model, we have to judge whether the
        module can be indexed directly.

        Args:
            model (nn.Module): This model may wrapped with DDP or not.
            module_name (str): The name of specific module.

        Return:
            nn.Module: Returned sub module.
        """
        module = model.module if hasattr(model, 'module') else model
        return getattr(module, module_name)

    def construct_fixed_noises(self):
        """Construct the fixed noises list used in SinGAN."""
        for i, real in enumerate(self.reals):
            h, w = real.shape[-2:]
            if i == 0:
                noise = torch.randn(1, 1, h, w).to(real)
                self.fixed_noises.append(noise)
            else:
                noise = torch.zeros_like(real)
                self.fixed_noises.append(noise)

    def forward(self,
                inputs: ForwardInputs,
                data_samples: Optional[list] = None,
                mode=None) -> List[DataSample]:
        """Forward function for SinGAN. For SinGAN, `inputs` should be a dict
        contains 'num_batches', 'mode' and other input arguments for the
        generator.

        Args:
            inputs (dict): Dict containing the necessary information
                (e.g., noise, num_batches, mode) to generate image.
            data_samples (Optional[list]): Data samples collated by
                :attr:`data_preprocessor`. Defaults to None.
            mode (Optional[str]): `mode` is not used in
                :class:`BaseConditionalGAN`. Defaults to None.
        """

        # handle batch_inputs
        assert isinstance(inputs, dict), (
            'SinGAN only support dict type inputs in forward function.')
        gen_kwargs = deepcopy(inputs)
        num_batches = gen_kwargs.pop('num_batches', 1)
        assert num_batches == 1, (
            'SinGAN only support \'num_batches\' as 1, but receive '
            f'{num_batches}.')
        sample_model = self._get_valid_model(inputs)
        gen_kwargs.pop('sample_model', None)  # remove sample_model

        mode = gen_kwargs.pop('mode', mode)
        mode = 'rand' if mode is None else mode
        curr_scale = gen_kwargs.pop('curr_scale', self.curr_stage)

        self.fixed_noises = [
            x.to(self.data_preprocessor.device) for x in self.fixed_noises
        ]

        batch_sample_list = []
        if sample_model in ['ema', 'orig']:
            if sample_model == 'ema':
                generator = self.generator_ema
            else:
                generator = self.generator

            outputs = generator(
                None,
                fixed_noises=self.fixed_noises,
                noise_weights=self.noise_weights,
                rand_mode=mode,
                num_batches=1,
                curr_scale=curr_scale,
                **gen_kwargs)

            gen_sample = DataSample()
            # destruct
            if isinstance(outputs, dict):
                outputs['fake_img'] = self.data_preprocessor.destruct(
                    outputs['fake_img'], data_samples)
                outputs['prev_res_list'] = [
                    self.data_preprocessor.destruct(r, data_samples)
                    for r in outputs['prev_res_list']
                ]
                gen_sample.fake_img = self.data_preprocessor.destruct(
                    outputs['fake_img'], data_samples)
                # gen_sample.prev_res_list = self.data_preprocessor.destruct(
                #     outputs['fake_img'], data_samples)
            else:
                outputs = self.data_preprocessor.destruct(
                    outputs, data_samples)

            # save to data sample
            for idx in range(num_batches):
                gen_sample = DataSample()
                # save inputs to data sample
                if data_samples:
                    gen_sample.update(data_samples[idx])
                if isinstance(outputs, dict):
                    gen_sample.fake_img = outputs['fake_img'][idx]
                    gen_sample.prev_res_list = [
                        r[idx] for r in outputs['prev_res_list']
                    ]
                else:
                    gen_sample.fake_img = outputs[idx]

                gen_sample.sample_model = sample_model
                batch_sample_list.append(gen_sample)

        else:  # sample model is 'ema/orig'

            outputs_orig = self.generator(
                None,
                fixed_noises=self.fixed_noises,
                noise_weights=self.noise_weights,
                rand_mode=mode,
                num_batches=1,
                curr_scale=curr_scale,
                **gen_kwargs)
            outputs_ema = self.generator_ema(
                None,
                fixed_noises=self.fixed_noises,
                noise_weights=self.noise_weights,
                rand_mode=mode,
                num_batches=1,
                curr_scale=curr_scale,
                **gen_kwargs)

            # destruct
            if isinstance(outputs_orig, dict):
                outputs_orig['fake_img'] = self.data_preprocessor.destruct(
                    outputs_orig['fake_img'], data_samples)
                outputs_orig['prev_res_list'] = [
                    self.data_preprocessor.destruct(r, data_samples)
                    for r in outputs_orig['prev_res_list']
                ]
                outputs_ema['fake_img'] = self.data_preprocessor.destruct(
                    outputs_ema['fake_img'], data_samples)
                outputs_ema['prev_res_list'] = [
                    self.data_preprocessor.destruct(r, data_samples)
                    for r in outputs_ema['prev_res_list']
                ]
            else:
                outputs_orig = self.data_preprocessor.destruct(
                    outputs_orig, data_samples)
                outputs_ema = self.data_preprocessor.destruct(
                    outputs_ema, data_samples)

            # save to data sample
            for idx in range(num_batches):
                gen_sample = DataSample()
                gen_sample.ema = DataSample()
                gen_sample.orig = DataSample()
                # save inputs to data sample
                if data_samples:
                    gen_sample.update(data_samples[idx])
                if isinstance(outputs_orig, dict):
                    gen_sample.ema.fake_img = outputs_ema['fake_img'][idx]
                    gen_sample.ema.prev_res_list = [
                        r[idx] for r in outputs_ema['prev_res_list']
                    ]
                    gen_sample.orig.fake_img = outputs_orig['fake_img'][idx]
                    gen_sample.orig.prev_res_list = [
                        r[idx] for r in outputs_orig['prev_res_list']
                    ]
                else:
                    gen_sample.ema.fake_img = outputs_ema[idx]
                    gen_sample.orig.fake_img = outputs_orig[idx]
                gen_sample.sample_model = sample_model

                batch_sample_list.append(gen_sample)
        return batch_sample_list

    def gen_loss(self, disc_pred_fake: Tensor,
                 recon_imgs: Tensor) -> Tuple[Tensor, dict]:
        r"""Generator loss for SinGAN. SinGAN use WGAN's loss and MSE loss to
        train the generator.

        .. math:
            L_{D} = -\mathbb{E}_{z\sim{p_{z}}}D\left\(G\left\(z\right\)\right\)
                + L_{MSE}
            L_{MSE} = \text{mean} \Vert x - G(z) \Vert_2

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            recon_imgs (Tensor): Reconstructive images.

        Returns:
            Tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_gen'] = -disc_pred_fake.mean()
        losses_dict['loss_mse'] = 10 * F.mse_loss(recon_imgs,
                                                  self.reals[self.curr_stage])
        loss, log_vars = self.parse_losses(losses_dict)
        return loss, log_vars

    def disc_loss(self, disc_pred_fake: Tensor, disc_pred_real: Tensor,
                  fake_data: Tensor, real_data: Tensor) -> Tuple[Tensor, dict]:
        r"""Get disc loss. SAGAN, SNGAN and Proj-GAN use hinge loss to train
        the generator.

        .. math:
            L_{D} = \mathbb{E}_{z\sim{p_{z}}}D\left\(G\left\(z\right\)\right\)
                - \mathbb{E}_{x\sim{p_{data}}}D\left\(x\right\) + L_{GP} \\
            L_{GP} = \lambda\mathbb{E}(\Vert\nabla_{\tilde{x}}D(\tilde{x})
                \Vert_2-1)^2 \\
            \tilde{x} = \epsilon x + (1-\epsilon)G(z)

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            disc_pred_real (Tensor): Discriminator's prediction of the real
                images.
            fake_data (Tensor): Generated images, used to calculate gradient
                penalty.
            real_data (Tensor): Real images, used to calculate gradient
                penalty.

        Returns:
            Tuple[Tensor, dict]: Loss value and a dict of log variables.
        """

        losses_dict = dict()
        losses_dict['loss_disc_fake'] = disc_pred_fake.mean()
        losses_dict['loss_disc_real'] = -disc_pred_real.mean()

        # gradient penalty
        batch_size = real_data.size(0)
        alpha = torch.rand(batch_size, 1, 1, 1).to(real_data)

        # interpolate between real_data and fake_data
        interpolates = alpha * real_data + (1. - alpha) * fake_data
        interpolates = autograd.Variable(interpolates, requires_grad=True)

        disc_interpolates = self.discriminator(
            interpolates, curr_scale=self.curr_stage)
        gradients = autograd.grad(
            outputs=disc_interpolates,
            inputs=interpolates,
            grad_outputs=torch.ones_like(disc_interpolates),
            create_graph=True,
            retain_graph=True,
            only_inputs=True)[0]
        # norm_mode is 'pixel'
        gradients_penalty = ((gradients.norm(2, dim=1) - 1)**2).mean()
        losses_dict['loss_gp'] = 0.1 * gradients_penalty

        parsed_loss, log_vars = self.parse_losses(losses_dict)
        return parsed_loss, log_vars

    def train_generator(self, inputs: dict, data_samples: List[DataSample],
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (List[DataSample]): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        fake_imgs = self.generator(
            inputs['input_sample'],
            self.fixed_noises,
            self.noise_weights,
            rand_mode='rand',
            curr_scale=self.curr_stage)
        disc_pred_fake_g = self.discriminator(
            fake_imgs, curr_scale=self.curr_stage)

        recon_imgs = self.generator(
            inputs['input_sample'],
            self.fixed_noises,
            self.noise_weights,
            rand_mode='recon',
            curr_scale=self.curr_stage)

        parsed_loss, log_vars = self.gen_loss(disc_pred_fake_g, recon_imgs)
        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def train_discriminator(self, inputs: dict, data_samples: List[DataSample],
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (List[DataSample]): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.
        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        input_sample = inputs['input_sample']
        fake_imgs = self.generator(
            input_sample,
            self.fixed_noises,
            self.noise_weights,
            rand_mode='rand',
            curr_scale=self.curr_stage)

        # disc pred for fake imgs and real_imgs
        real_imgs = self.reals[self.curr_stage]
        disc_pred_fake = self.discriminator(fake_imgs.detach(),
                                            self.curr_stage)
        disc_pred_real = self.discriminator(real_imgs, self.curr_stage)
        parsed_loss, log_vars = self.disc_loss(disc_pred_fake, disc_pred_real,
                                               fake_imgs, real_imgs)
        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def train_gan(self, inputs_dict: dict, data_sample: List[DataSample],
                  optim_wrapper: OptimWrapperDict) -> Dict[str, torch.Tensor]:
        """Train GAN model. In the training of GAN models, generator and
        discriminator are updated alternatively. In MMagic's design,
        `self.train_step` is called with data input. Therefore we always update
        discriminator, whose updating is relay on real data, and then determine
        if the generator needs to be updated based on the current number of
        iterations. More details about whether to update generator can be found
        in :meth:`should_gen_update`.

        Args:
            data (dict): Data sampled from dataloader.
            data_sample (List[DataSample]): List of data sample contains GT
                and meta information.
            optim_wrapper (OptimWrapperDict): OptimWrapperDict instance
                contains OptimWrapper of generator and discriminator.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """
        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')

        disc_optimizer_wrapper: OptimWrapper = optim_wrapper['discriminator']
        disc_accu_iters = disc_optimizer_wrapper._accumulative_counts

        with disc_optimizer_wrapper.optim_context(self.discriminator):
            log_vars = self.train_discriminator(inputs_dict, data_sample,
                                                disc_optimizer_wrapper)

        # add 1 to `curr_iter` because iter is updated in train loop.
        # Whether to update the generator. We update generator with
        # discriminator is fully updated for `self.n_discriminator_steps`
        # iterations. And one full updating for discriminator contains
        # `disc_accu_counts` times of grad accumulations.
        if (curr_iter + 1) % (self.discriminator_steps * disc_accu_iters) == 0:
            set_requires_grad(self.discriminator, False)
            gen_optimizer_wrapper = optim_wrapper['generator']
            gen_accu_iters = gen_optimizer_wrapper._accumulative_counts

            log_vars_gen_list = []
            # init optimizer wrapper status for generator manually
            gen_optimizer_wrapper.initialize_count_status(
                self.generator, 0, self.generator_steps * gen_accu_iters)
            for _ in range(self.generator_steps * gen_accu_iters):
                with gen_optimizer_wrapper.optim_context(self.generator):
                    log_vars_gen = self.train_generator(
                        inputs_dict, data_sample, gen_optimizer_wrapper)

                log_vars_gen_list.append(log_vars_gen)
            log_vars_gen = self.gather_log_vars(log_vars_gen_list)
            log_vars_gen.pop('loss', None)  # remove 'loss' from gen logs

            set_requires_grad(self.discriminator, True)

            # only do ema after generator update
            if self.with_ema_gen and (curr_iter + 1) >= (
                    self.ema_start * self.discriminator_steps *
                    disc_accu_iters):
                self.generator_ema.update_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)
                # if not update buffer, copy buffer from orig model
                if not self.generator_ema.update_buffers:
                    self.generator_ema.sync_buffers(
                        self.generator.module if is_model_wrapper(
                            self.generator) else self.generator)
            elif self.with_ema_gen:
                # before ema, copy weights from orig
                self.generator_ema.sync_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)

            log_vars.update(log_vars_gen)
        return log_vars

    def train_step(self, data: dict,
                   optim_wrapper: OptimWrapperDict) -> Dict[str, Tensor]:
        """Train step for SinGAN model. SinGAN is trained with multi-resolution
        images, and each resolution is trained for `:attr:self.iters_per_scale`
        times.

        We initialize the weight and learning rate scheduler of the
        corresponding module at the start of each resolution's training. At
        the end of each resolution's training, we update the weight of the
        noise of current resolution by mse loss between reconstructed image and
        real image.

        Args:
            data (dict): Data sampled from dataloader.
            optim_wrapper (OptimWrapperDict): OptimWrapperDict instance
                contains OptimWrapper of generator and discriminator.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """

        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')

        if curr_iter % (self.iters_per_scale * self.discriminator_steps) == 0:
            self.curr_stage += 1
            # load weights from prev scale
            self.get_module(self.generator, 'check_and_load_prev_weight')(
                self.curr_stage)
            self.get_module(self.discriminator, 'check_and_load_prev_weight')(
                self.curr_stage)

            # assert grad_accumulation step is 1
            curr_gen_optim = optim_wrapper[f'generator_{self.curr_stage}']
            curr_disc_optim = optim_wrapper[f'discriminator_{self.curr_stage}']
            _warning_msg = ('SinGAN do set batch size as 1 during training '
                            'and do not support gradient accumulation.')
            assert curr_gen_optim._accumulative_counts == 1, _warning_msg
            assert curr_disc_optim._accumulative_counts == 1, _warning_msg

            # build parameters scheduler manually, because parameters_schedule
            # hook update all scheduler at the same
            if self.lr_scheduler_args:
                self.gen_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                    curr_gen_optim.optimizer, **self.lr_scheduler_args)
                self.disc_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                    curr_disc_optim.optimizer, **self.lr_scheduler_args)

        # get current optimizer_wrapper
        curr_optimizer_wrapper = OptimWrapperDict(
            generator=optim_wrapper[f'generator_{self.curr_stage}'],
            discriminator=optim_wrapper[f'discriminator_{self.curr_stage}'])

        # handle inputs
        data = self.data_preprocessor(data)
        inputs_dict, data_samples = data['inputs'], data['data_samples']
        # setup fixed noises and reals pyramid
        if curr_iter == 0 or len(self.reals) == 0:
            keys = [k for k in inputs_dict.keys() if 'real_scale' in k]
            scales = len(keys)
            self.reals = [inputs_dict[f'real_scale{s}'] for s in range(scales)]

            # here we do not padding fixed noises
            self.construct_fixed_noises()

        # standard train step
        log_vars = self.train_gan(inputs_dict, data_samples,
                                  curr_optimizer_wrapper)
        log_vars['curr_stage'] = self.curr_stage

        # update noise weight
        if ((curr_iter + 1) % (self.iters_per_scale * self.discriminator_steps)
                == 0) and (self.curr_stage < len(self.reals) - 1):
            with torch.no_grad():
                g_recon = self.generator(
                    inputs_dict['input_sample'],
                    self.fixed_noises,
                    self.noise_weights,
                    rand_mode='recon',
                    curr_scale=self.curr_stage)
                if isinstance(g_recon, dict):
                    g_recon = g_recon['fake_img']
                g_recon = F.interpolate(
                    g_recon, self.reals[self.curr_stage + 1].shape[-2:])

            mse = F.mse_loss(g_recon.detach(), self.reals[self.curr_stage + 1])
            rmse = torch.sqrt(mse)
            self.noise_weights.append(self.noise_weight_init * rmse.item())

        # call scheduler when all submodules are fully updated.
        if (curr_iter + 1) % self.discriminator_steps == 0:
            if self.lr_scheduler_args:
                self.disc_scheduler.step()
                self.gen_scheduler.step()

        return log_vars

    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data in test progress. Before
        generate images, we call `:meth:self.load_test_pkl` to load the fixed
        noise and current stage of the model from the pickle file.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            SampleList: A list of ``DataSample`` contain generated results.
        """
        if not self.loaded_test_pkl:
            self.load_test_pkl()
        return super().test_step(data)



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy

import torch
import torch.nn as nn
import torch.nn.functional as F

try:
    from mmcv.ops import conv2d, conv_transpose2d
except ImportError:
    conv2d = None
    conv_transpose2d = None
    print('Warning: mmcv.ops.conv2d, mmcv.ops.conv_transpose2d'
          'are not available.')

from ..pggan import equalized_lr
from ..stylegan1 import Blur, EqualLinearActModule, NoiseInjection
from ..stylegan2.stylegan2_modules import _FusedBiasLeakyReLU


class ModulatedPEConv2d(nn.Module):
    r"""Modulated Conv2d in StyleGANv2 with Positional Encoding (PE).

    This module is modified from the ``ModulatedConv2d`` in StyleGAN2 to
    support the experiments in: Positional Encoding as Spatial Inductive Bias
    in GANs, CVPR'2021.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        kernel_size (int): Kernel size, same as :obj:`nn.Con2d`.
        style_channels (int): Channels for the style codes.
        demodulate (bool, optional): Whether to adopt demodulation.
            Defaults to True.
        upsample (bool, optional): Whether to adopt upsampling in features.
            Defaults to False.
        downsample (bool, optional): Whether to adopt downsampling in features.
            Defaults to False.
        blur_kernel (list[int], optional): Blurry kernel.
            Defaults to [1, 3, 3, 1].
        equalized_lr_cfg (dict | None, optional): Configs for equalized lr.
            Defaults to dict(mode='fan_in', lr_mul=1., gain=1.).
        style_mod_cfg (dict, optional): Configs for style modulation module.
            Defaults to dict(bias_init=1.).
        style_bias (float, optional): Bias value for style code.
            Defaults to 0..
        eps (float, optional): Epsilon value to avoid computation error.
            Defaults to 1e-8.
        no_pad (bool, optional): Whether to removing the padding in
            convolution. Defaults to False.
        deconv2conv (bool, optional): Whether to substitute the transposed conv
            with (conv2d, upsampling). Defaults to False.
        interp_pad (int | None, optional): The padding number of interpolation
            pad. Defaults to None.
        up_config (dict, optional): Upsampling config.
            Defaults to dict(scale_factor=2, mode='nearest').
        up_after_conv (bool, optional): Whether to adopt upsampling after
            convolution. Defaults to False.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 style_channels,
                 demodulate=True,
                 upsample=False,
                 downsample=False,
                 blur_kernel=[1, 3, 3, 1],
                 equalized_lr_cfg=dict(mode='fan_in', lr_mul=1., gain=1.),
                 style_mod_cfg=dict(bias_init=1.),
                 style_bias=0.,
                 eps=1e-8,
                 no_pad=False,
                 deconv2conv=False,
                 interp_pad=None,
                 up_config=dict(scale_factor=2, mode='nearest'),
                 up_after_conv=False):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.style_channels = style_channels
        self.demodulate = demodulate
        # sanity check for kernel size
        assert isinstance(self.kernel_size,
                          int) and (self.kernel_size >= 1
                                    and self.kernel_size % 2 == 1)
        self.upsample = upsample
        self.downsample = downsample
        self.style_bias = style_bias
        self.eps = eps
        self.no_pad = no_pad
        self.deconv2conv = deconv2conv
        self.interp_pad = interp_pad
        self.with_interp_pad = interp_pad is not None
        self.up_config = deepcopy(up_config)
        self.up_after_conv = up_after_conv

        # build style modulation module
        style_mod_cfg = dict() if style_mod_cfg is None else style_mod_cfg

        self.style_modulation = EqualLinearActModule(style_channels,
                                                     in_channels,
                                                     **style_mod_cfg)
        # set lr_mul for conv weight
        lr_mul_ = 1.
        if equalized_lr_cfg is not None:
            lr_mul_ = equalized_lr_cfg.get('lr_mul', 1.)
        self.weight = nn.Parameter(
            torch.randn(1, out_channels, in_channels, kernel_size,
                        kernel_size).div_(lr_mul_))

        # build blurry layer for upsampling
        if upsample and not self.deconv2conv:
            factor = 2
            p = (len(blur_kernel) - factor) - (kernel_size - 1)
            pad0 = (p + 1) // 2 + factor - 1
            pad1 = p // 2 + 1
            self.blur = Blur(blur_kernel, (pad0, pad1), upsample_factor=factor)

        # build blurry layer for downsampling
        if downsample:
            factor = 2
            p = (len(blur_kernel) - factor) + (kernel_size - 1)
            pad0 = (p + 1) // 2
            pad1 = p // 2
            self.blur = Blur(blur_kernel, pad=(pad0, pad1))

        # add equalized_lr hook for conv weight
        if equalized_lr_cfg is not None:
            equalized_lr(self, **equalized_lr_cfg)

        # if `no_pad`, remove all of the padding in conv
        self.padding = kernel_size // 2 if not no_pad else 0

    def forward(self, x, style):
        """Forward function.

        Args:
            x ([Tensor): Input features with shape of (N, C, H, W).
            style (Tensor): Style latent with shape of (N, C).

        Returns:
            Tensor: Output feature with shape of (N, C, H, W).
        """
        n, c, h, w = x.shape
        # process style code
        style = self.style_modulation(style).view(n, 1, c, 1,
                                                  1) + self.style_bias

        # combine weight and style
        weight = self.weight * style
        if self.demodulate:
            demod = torch.rsqrt(weight.pow(2).sum([2, 3, 4]) + self.eps)
            weight = weight * demod.view(n, self.out_channels, 1, 1, 1)

        weight = weight.view(n * self.out_channels, c, self.kernel_size,
                             self.kernel_size)

        if self.upsample and not self.deconv2conv:
            x = x.reshape(1, n * c, h, w)
            weight = weight.view(n, self.out_channels, c, self.kernel_size,
                                 self.kernel_size)
            weight = weight.transpose(1, 2).reshape(n * c, self.out_channels,
                                                    self.kernel_size,
                                                    self.kernel_size)
            x = conv_transpose2d(x, weight, padding=0, stride=2, groups=n)
            x = x.reshape(n, self.out_channels, *x.shape[-2:])
            x = self.blur(x)
        elif self.upsample and self.deconv2conv:
            if self.up_after_conv:
                x = x.reshape(1, n * c, h, w)
                x = conv2d(x, weight, padding=self.padding, groups=n)
                x = x.view(n, self.out_channels, *x.shape[2:4])

            if self.with_interp_pad:
                h_, w_ = x.shape[-2:]
                up_cfg_ = deepcopy(self.up_config)
                up_scale = up_cfg_.pop('scale_factor')
                size_ = (h_ * up_scale + self.interp_pad,
                         w_ * up_scale + self.interp_pad)
                x = F.interpolate(x, size=size_, **up_cfg_)
            else:
                x = F.interpolate(x, **self.up_config)

            if not self.up_after_conv:
                h_, w_ = x.shape[-2:]
                x = x.view(1, n * c, h_, w_)
                x = conv2d(x, weight, padding=self.padding, groups=n)
                x = x.view(n, self.out_channels, *x.shape[2:4])

        elif self.downsample:
            x = self.blur(x)
            x = x.view(1, n * self.in_channels, *x.shape[-2:])
            x = conv2d(x, weight, stride=2, padding=0, groups=n)
            x = x.view(n, self.out_channels, *x.shape[-2:])
        else:
            x = x.view(1, n * c, h, w)
            x = conv2d(x, weight, stride=1, padding=self.padding, groups=n)
            x = x.view(n, self.out_channels, *x.shape[-2:])

        return x


class ModulatedPEStyleConv(nn.Module):
    """Modulated Style Convolution with Positional Encoding.

    This module is modified from the ``ModulatedStyleConv`` in StyleGAN2 to
    support the experiments in: Positional Encoding as Spatial Inductive Bias
    in GANs, CVPR'2021.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        kernel_size (int): Kernel size, same as :obj:`nn.Con2d`.
        style_channels (int): Channels for the style codes.
        demodulate (bool, optional): Whether to adopt demodulation.
            Defaults to True.
        upsample (bool, optional): Whether to adopt upsampling in features.
            Defaults to False.
        downsample (bool, optional): Whether to adopt downsampling in features.
            Defaults to False.
        blur_kernel (list[int], optional): Blurry kernel.
            Defaults to [1, 3, 3, 1].
        equalized_lr_cfg (dict | None, optional): Configs for equalized lr.
            Defaults to dict(mode='fan_in', lr_mul=1., gain=1.).
        style_mod_cfg (dict, optional): Configs for style modulation module.
            Defaults to dict(bias_init=1.).
        style_bias (float, optional): Bias value for style code.
            Defaults to 0..
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 style_channels,
                 upsample=False,
                 blur_kernel=[1, 3, 3, 1],
                 demodulate=True,
                 style_mod_cfg=dict(bias_init=1.),
                 style_bias=0.,
                 **kwargs):
        super().__init__()

        self.conv = ModulatedPEConv2d(
            in_channels,
            out_channels,
            kernel_size,
            style_channels,
            demodulate=demodulate,
            upsample=upsample,
            blur_kernel=blur_kernel,
            style_mod_cfg=style_mod_cfg,
            style_bias=style_bias,
            **kwargs)

        self.noise_injector = NoiseInjection()
        self.activate = _FusedBiasLeakyReLU(out_channels)

    def forward(self, x, style, noise=None, return_noise=False):
        """Forward Function.

        Args:
            x ([Tensor): Input features with shape of (N, C, H, W).
            style (Tensor): Style latent with shape of (N, C).
            noise (Tensor, optional): Noise for injection. Defaults to None.
            return_noise (bool, optional): Whether to return noise tensors.
                Defaults to False.

        Returns:
            Tensor: Output features with shape of (N, C, H, W)
        """
        out = self.conv(x, style)
        if return_noise:
            out, noise = self.noise_injector(
                out, noise=noise, return_noise=return_noise)
        else:
            out = self.noise_injector(
                out, noise=noise, return_noise=return_noise)

        out = self.activate(out)

        if return_noise:
            return out, noise

        return out



# Copyright (c) OpenMMLab. All rights reserved.
from .liif import LIIF
from .liif_net import LIIFEDSRNet, LIIFRDNNet
from .mlp_refiner import MLPRefiner

__all__ = [
    'LIIF',
    'LIIFEDSRNet',
    'LIIFRDNNet',
    'MLPRefiner',
]



# Copyright (c) OpenMMLab. All rights reserved.
import re
from typing import Tuple, Union

import torch.nn as nn
from mmengine import print_log
from mmengine.optim import (DefaultOptimWrapperConstructor, OptimWrapper,
                            OptimWrapperDict)

from mmagic.registry import (OPTIM_WRAPPER_CONSTRUCTORS, OPTIM_WRAPPERS,
                             OPTIMIZERS)


@OPTIM_WRAPPER_CONSTRUCTORS.register_module()
class MultiOptimWrapperConstructor:
    """OptimizerConstructor for GAN models. This class construct optimizer for
    the submodules of the model separately, and return a
    :class:`mmengine.optim.OptimWrapperDict` or
    :class:`mmengine.optim.OptimWrapper`.

    Example 1: Build multi optimizers (e.g., GANs):
        >>> # build GAN model
        >>> model = dict(
        >>>     type='GANModel',
        >>>     num_classes=10,
        >>>     generator=dict(type='Generator'),
        >>>     discriminator=dict(type='Discriminator'))
        >>> gan_model = MODELS.build(model)
        >>> # build constructor
        >>> optim_wrapper = dict(
        >>>     generator=dict(
        >>>         type='OptimWrapper',
        >>>         accumulative_counts=1,
        >>>         optimizer=dict(type='Adam', lr=0.0002,
        >>>                        betas=(0.5, 0.999))),
        >>>     discriminator=dict(
        >>>         type='OptimWrapper',
        >>>         accumulative_counts=1,
        >>>         optimizer=dict(type='Adam', lr=0.0002,
        >>>                            betas=(0.5, 0.999))))
        >>> optim_dict_builder = MultiOptimWrapperConstructor(optim_wrapper)
        >>> # build optim wrapper dict
        >>> optim_wrapper_dict = optim_dict_builder(gan_model)

    Example 2: Build multi optimizers for specific submodules:
        >>> # build model
        >>> class GAN(nn.Module):
        >>>     def __init__(self) -> None:
        >>>         super().__init__()
        >>>         self.generator = nn.Conv2d(3, 3, 1)
        >>>         self.discriminator = nn.Conv2d(3, 3, 1)
        >>> class TextEncoder(nn.Module):
        >>>     def __init__(self):
        >>>         super().__init__()
        >>>         self.embedding = nn.Embedding(100, 100)
        >>> class ToyModel(nn.Module):
        >>>     def __init__(self) -> None:
        >>>         super().__init__()
        >>>         self.m1 = GAN()
        >>>         self.m2 = nn.Conv2d(3, 3, 1)
        >>>         self.m3 = nn.Linear(2, 2)
        >>>         self.text_encoder = TextEncoder()
        >>> model = ToyModel()
        >>> # build constructor
        >>> optim_wrapper = {
        >>>     '.*embedding': {
        >>>         'type': 'OptimWrapper',
        >>>         'optimizer': {
        >>>             'type': 'Adam',
        >>>             'lr': 1e-4,
        >>>             'betas': (0.9, 0.99)
        >>>         }
        >>>     },
        >>>     'm1.generator': {
        >>>         'type': 'OptimWrapper',
        >>>         'optimizer': {
        >>>             'type': 'Adam',
        >>>             'lr': 1e-5,
        >>>             'betas': (0.9, 0.99)
        >>>         }
        >>>     },
        >>>     'm2': {
        >>>         'type': 'OptimWrapper',
        >>>         'optimizer': {
        >>>             'type': 'Adam',
        >>>             'lr': 1e-5,
        >>>         }
        >>>     }
        >>> }
        >>> optim_dict_builder = MultiOptimWrapperConstructor(optim_wrapper)
        >>> # build optim wrapper dict
        >>> optim_wrapper_dict = optim_dict_builder(model)

    Example 3: Build a single optimizer for multi modules (e.g., DreamBooth):
        >>> # build StableDiffusion model
        >>> model = dict(
        >>>     type='StableDiffusion',
        >>>     unet=dict(type='unet'),
        >>>     vae=dict(type='vae'),
                text_encoder=dict(type='text_encoder'))
        >>> diffusion_model = MODELS.build(model)
        >>> # build constructor
        >>> optim_wrapper = dict(
        >>>     modules=['unet', 'text_encoder']
        >>>     optimizer=dict(type='Adam', lr=0.0002),
        >>>     accumulative_counts=1)
        >>> optim_dict_builder = MultiOptimWrapperConstructor(optim_wrapper)
        >>> # build optim wrapper dict
        >>> optim_wrapper_dict = optim_dict_builder(diffusion_model)

    Args:
        optim_wrapper_cfg_dict (dict): Config of the optimizer wrapper.
        paramwise_cfg (dict): Config of parameter-wise settings. Default: None.
    """

    def __init__(self, optim_wrapper_cfg: dict, paramwise_cfg=None):

        if not isinstance(optim_wrapper_cfg, dict):
            raise TypeError('optimizer_cfg should be a dict',
                            f'but got {type(optim_wrapper_cfg)}')
        assert paramwise_cfg is None, (
            'paramwise_cfg should be set in each optimizer separately')
        self.optim_cfg = optim_wrapper_cfg

        if 'modules' in optim_wrapper_cfg:
            # single optimizer with multi param groups
            cfg_ = optim_wrapper_cfg.copy()
            self.modules = cfg_.pop('modules')
            paramwise_cfg_ = cfg_.pop('paramwise_cfg', None)
            self.constructors = DefaultOptimWrapperConstructor(
                cfg_, paramwise_cfg_)
        else:
            self.constructors = {}
            self.modules = {}
            for key, cfg in self.optim_cfg.items():
                cfg_ = cfg.copy()
                if 'modules' in cfg_:
                    self.modules[key] = cfg_.pop('modules')
                paramwise_cfg_ = cfg_.pop('paramwise_cfg', None)
                self.constructors[key] = DefaultOptimWrapperConstructor(
                    cfg_, paramwise_cfg_)

    def __call__(self,
                 module: nn.Module) -> Union[OptimWrapperDict, OptimWrapper]:
        """Build optimizer and return a optimizer_wrapper_dict."""

        optimizers = {}
        if hasattr(module, 'module'):
            module = module.module
        if isinstance(self.constructors, dict):
            for key, constructor in self.constructors.items():
                module_names = self.modules[key] if self.modules else key
                if (isinstance(module_names, str)
                        and module_names in module._modules):
                    optimizers[key] = constructor(
                        module._modules[module_names])
                    optim_wrapper_cfg = constructor.optimizer_cfg
                    print_log(
                        f'Add to optimizer \'{key}\' '
                        f'({optim_wrapper_cfg}): \'{key}\'.', 'current')
                else:

                    assert not constructor.paramwise_cfg, (
                        'Do not support paramwise_cfg for multi module '
                        'optimizer.')

                    params, found_names = get_params_by_names(
                        module, module_names)
                    # build optimizer
                    optimizer_cfg = constructor.optimizer_cfg.copy()
                    optimizer_cfg['params'] = params
                    optimizer = OPTIMIZERS.build(optimizer_cfg)

                    # build optimizer wrapper
                    optim_wrapper_cfg = constructor.optim_wrapper_cfg.copy()
                    optim_wrapper_cfg.setdefault('type', 'OptimWrapper')
                    optim_wrapper = OPTIM_WRAPPERS.build(
                        optim_wrapper_cfg,
                        default_args=dict(optimizer=optimizer))

                    for name in found_names:
                        print_log(
                            f'Add to optimizer \'{key}\' '
                            f'({constructor.optimizer_cfg}): \'{name}\'.',
                            'current')

                    optimizers[key] = optim_wrapper

            return OptimWrapperDict(**optimizers)

        else:
            params, found_names = get_params_by_names(module, self.modules)

            constructor = self.constructors
            assert not constructor.paramwise_cfg, (
                'Do not support paramwise_cfg for multi parameters')

            optimizer_cfg = constructor.optimizer_cfg.copy()
            optimizer_cfg['params'] = params
            optimizer = OPTIMIZERS.build(optimizer_cfg)
            for name in found_names:
                print_log(
                    f'Add to optimizer ({constructor.optimizer_cfg}): '
                    f'\'{name}\'.', 'current')

            # build optimizer wrapper
            optim_wrapper_cfg = constructor.optim_wrapper_cfg.copy()
            optim_wrapper_cfg.setdefault('type', 'OptimWrapper')
            optim_wrapper = OPTIM_WRAPPERS.build(
                optim_wrapper_cfg, default_args=dict(optimizer=optimizer))

            return optim_wrapper


def get_params_by_names(module: nn.Module,
                        names: Union[str, list]) -> Tuple[list, list]:
    """Support two kinds of name matching:
        1. matching name from **first-level** submodule.
        2. matching name by `re.fullmatch`.

    Args:
        module (nn.Module): The module to get parameters.
        names (Union[str, list]): The name or a list of names of the
            submodule parameters.

    Returns:
        Tuple[list]: A list of parameters and corresponding name for logging.
    """

    if not isinstance(names, list):
        names = [names]

    params = []
    found_names = []
    for name in names:
        if name in module._modules:
            params.extend(module._modules[name].parameters())
            found_names.append(name)
        else:
            for n, m in module.named_modules():
                if re.fullmatch(name, n):
                    params.extend(m.parameters())
                    found_names.append(n)
    return params, found_names



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch.nn as nn
from mmengine.optim import DefaultOptimWrapperConstructor, OptimWrapperDict

from mmagic.registry import OPTIM_WRAPPER_CONSTRUCTORS


@OPTIM_WRAPPER_CONSTRUCTORS.register_module()
class SinGANOptimWrapperConstructor:
    """OptimizerConstructor for SinGAN models. Set optimizers for each
    submodule of SinGAN. All submodule must be contained in a
    :class:`torch.nn.ModuleList` named 'blocks'. And we access each submodule
    by `MODEL.blocks[SCALE]`, where `MODEL` is generator or discriminator, and
    the scale is the index of the resolution scale.

    More detail about the resolution scale and naming rule please refers to
    :class:`~mmagic.models.editors.singan.SinGANMultiScaleGenerator` and
    :class:`~mmagic.models.editors.singan.SinGANMultiScaleDiscriminator`.

    Example:
        >>> # build SinGAN model
        >>> model = dict(
        >>>     type='SinGAN',
        >>>     data_preprocessor=dict(
        >>>         type='GANDataPreprocessor',
        >>>         non_image_keys=['input_sample']),
        >>>     generator=dict(
        >>>         type='SinGANMultiScaleGenerator',
        >>>         in_channels=3,
        >>>         out_channels=3,
        >>>         num_scales=2),
        >>>     discriminator=dict(
        >>>         type='SinGANMultiScaleDiscriminator',
        >>>         in_channels=3,
        >>>         num_scales=3))
        >>> singan = MODELS.build(model)
        >>> # build constructor
        >>> optim_wrapper = dict(
        >>>     generator=dict(optimizer=dict(type='Adam', lr=0.0005,
        >>>                                   betas=(0.5, 0.999))),
        >>>     discriminator=dict(
        >>>         optimizer=dict(type='Adam', lr=0.0005,
        >>>                        betas=(0.5, 0.999))))
        >>> optim_wrapper_dict_builder = SinGANOptimWrapperConstructor(
        >>>     optim_wrapper)
        >>> # build optim wrapper dict
        >>> optim_wrapper_dict = optim_wrapper_dict_builder(singan)

    Args:
        optim_wrapper_cfg (dict): Config of the optimizer wrapper.
        paramwise_cfg (Optional[dict]): Parameter-wise options.
    """

    def __init__(self,
                 optim_wrapper_cfg: dict,
                 paramwise_cfg: Optional[dict] = None):
        if not isinstance(optim_wrapper_cfg, dict):
            raise TypeError('optimizer_cfg should be a dict',
                            f'but got {type(optim_wrapper_cfg)}')
        assert paramwise_cfg is None, (
            'paramwise_cfg should be set in each optimizer separately')
        self.optim_cfg = optim_wrapper_cfg
        self.constructors = {}
        for key, cfg in self.optim_cfg.items():
            cfg_ = cfg.copy()
            paramwise_cfg_ = cfg_.pop('paramwise_cfg', None)
            self.constructors[key] = DefaultOptimWrapperConstructor(
                cfg_, paramwise_cfg_)

    def __call__(self, module: nn.Module) -> OptimWrapperDict:
        """Build optimizer and return a optimizerwrapperdict."""
        optimizers = {}
        if hasattr(module, 'module'):
            module = module.module
        num_scales = module.num_scales

        for key, constructor in self.constructors.items():
            for idx in range(num_scales + 1):
                submodule = module._modules[key]
                if hasattr(submodule, 'module'):
                    submodule = submodule.module
                optimizers[f'{key}_{idx}'] = constructor(submodule.blocks[idx])
        optimizers = OptimWrapperDict(**optimizers)
        return optimizers



_base_ = [
    '../_base_/models/dcgan/base_dcgan_64x64.py',
    '../_base_/datasets/unconditional_imgs_64x64.py',
    '../_base_/gen_default_runtime.py'
]

# output single channel
model = dict(
    data_preprocessor=dict(mean=[127.5], std=[127.5]),
    generator=dict(out_channels=1),
    discriminator=dict(in_channels=1))

# define dataset
# modify train_pipeline to load gray scale images
train_pipeline = [
    dict(type='LoadImageFromFile', key='gt', color_type='grayscale'),
    dict(type='Resize', keys='gt', scale=(64, 64)),
    dict(type='PackInputs')
]

# set ``batch_size``` and ``data_root```
batch_size = 128
data_root = 'data/mnist_64/train'
train_dataloader = dict(
    batch_size=batch_size,
    dataset=dict(data_root=data_root, pipeline=train_pipeline))

val_dataloader = dict(
    batch_size=batch_size,
    dataset=dict(data_root=data_root, pipeline=train_pipeline))

test_dataloader = dict(
    batch_size=batch_size,
    dataset=dict(data_root=data_root, pipeline=train_pipeline))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=500,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]

train_cfg = dict(max_iters=5000, val_interval=500)

# METRICS
metrics = [
    dict(
        type='MS_SSIM', prefix='ms-ssim', fake_nums=10000,
        sample_model='orig'),
    dict(
        type='SWD',
        prefix='swd',
        fake_nums=-1,
        sample_model='orig',
        image_shape=(1, 64, 64))
]
# save best checkpoints
default_hooks = dict(
    checkpoint=dict(interval=500, save_best='swd/avg', rule='less'))

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0004, betas=(0.5, 0.999))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.999))))



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import scipy.signal
import torch

try:
    from mmcv.ops import conv2d
except ImportError:
    conv2d = None
    print('Warning: mmcv.ops.conv2d are not available.')

from . import grid_sample_gradfix, misc, upfirdn2d

# ----------------------------------------------------------------------------
# Coefficients of various wavelet decomposition low-pass filters.

wavelets = {
    'haar': [0.7071067811865476, 0.7071067811865476],
    'db1': [0.7071067811865476, 0.7071067811865476],
    'db2': [
        -0.12940952255092145, 0.22414386804185735, 0.836516303737469,
        0.48296291314469025
    ],
    'db3': [
        0.035226291882100656, -0.08544127388224149, -0.13501102001039084,
        0.4598775021193313, 0.8068915093133388, 0.3326705529509569
    ],
    'db4': [
        -0.010597401784997278, 0.032883011666982945, 0.030841381835986965,
        -0.18703481171888114, -0.02798376941698385, 0.6308807679295904,
        0.7148465705525415, 0.23037781330885523
    ],
    'db5': [
        0.003335725285001549, -0.012580751999015526, -0.006241490213011705,
        0.07757149384006515, -0.03224486958502952, -0.24229488706619015,
        0.13842814590110342, 0.7243085284385744, 0.6038292697974729,
        0.160102397974125
    ],
    'db6': [
        -0.00107730108499558, 0.004777257511010651, 0.0005538422009938016,
        -0.031582039318031156, 0.02752286553001629, 0.09750160558707936,
        -0.12976686756709563, -0.22626469396516913, 0.3152503517092432,
        0.7511339080215775, 0.4946238903983854, 0.11154074335008017
    ],
    'db7': [
        0.0003537138000010399, -0.0018016407039998328, 0.00042957797300470274,
        0.012550998556013784, -0.01657454163101562, -0.03802993693503463,
        0.0806126091510659, 0.07130921926705004, -0.22403618499416572,
        -0.14390600392910627, 0.4697822874053586, 0.7291320908465551,
        0.39653931948230575, 0.07785205408506236
    ],
    'db8': [
        -0.00011747678400228192, 0.0006754494059985568, -0.0003917403729959771,
        -0.00487035299301066, 0.008746094047015655, 0.013981027917015516,
        -0.04408825393106472, -0.01736930100202211, 0.128747426620186,
        0.00047248457399797254, -0.2840155429624281, -0.015829105256023893,
        0.5853546836548691, 0.6756307362980128, 0.3128715909144659,
        0.05441584224308161
    ],
    'sym2': [
        -0.12940952255092145, 0.22414386804185735, 0.836516303737469,
        0.48296291314469025
    ],
    'sym3': [
        0.035226291882100656, -0.08544127388224149, -0.13501102001039084,
        0.4598775021193313, 0.8068915093133388, 0.3326705529509569
    ],
    'sym4': [
        -0.07576571478927333, -0.02963552764599851, 0.49761866763201545,
        0.8037387518059161, 0.29785779560527736, -0.09921954357684722,
        -0.012603967262037833, 0.0322231006040427
    ],
    'sym5': [
        0.027333068345077982, 0.029519490925774643, -0.039134249302383094,
        0.1993975339773936, 0.7234076904024206, 0.6339789634582119,
        0.01660210576452232, -0.17532808990845047, -0.021101834024758855,
        0.019538882735286728
    ],
    'sym6': [
        0.015404109327027373, 0.0034907120842174702, -0.11799011114819057,
        -0.048311742585633, 0.4910559419267466, 0.787641141030194,
        0.3379294217276218, -0.07263752278646252, -0.021060292512300564,
        0.04472490177066578, 0.0017677118642428036, -0.007800708325034148
    ],
    'sym7': [
        0.002681814568257878, -0.0010473848886829163, -0.01263630340325193,
        0.03051551316596357, 0.0678926935013727, -0.049552834937127255,
        0.017441255086855827, 0.5361019170917628, 0.767764317003164,
        0.2886296317515146, -0.14004724044296152, -0.10780823770381774,
        0.004010244871533663, 0.010268176708511255
    ],
    'sym8': [
        -0.0033824159510061256, -0.0005421323317911481, 0.03169508781149298,
        0.007607487324917605, -0.1432942383508097, -0.061273359067658524,
        0.4813596512583722, 0.7771857517005235, 0.3644418948353314,
        -0.05194583810770904, -0.027219029917056003, 0.049137179673607506,
        0.003808752013890615, -0.01495225833704823, -0.0003029205147213668,
        0.0018899503327594609
    ],
}

# ----------------------------------------------------------------------------
# Helpers for constructing transformation matrices.


def matrix(*rows, device=None):
    """Constructing transformation matrices.

    Args:
        device (str|torch.device, optional): Matrix device. Defaults to None.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    assert all(len(row) == len(rows[0]) for row in rows)
    elems = [x for row in rows for x in row]
    ref = [x for x in elems if isinstance(x, torch.Tensor)]
    if len(ref) == 0:
        return misc.constant(np.asarray(rows), device=device)
    assert device is None or device == ref[0].device
    # change `x.float()` to support pt1.5
    elems = [
        x.float() if isinstance(x, torch.Tensor) else misc.constant(
            x, shape=ref[0].shape, device=ref[0].device) for x in elems
    ]
    return torch.stack(elems, dim=-1).reshape(ref[0].shape + (len(rows), -1))


def translate2d(tx, ty, **kwargs):
    """Construct 2d translation matrix.

    Args:
        tx (float): X-direction translation amount.
        ty (float): Y-direction translation amount.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return matrix([1, 0, tx], [0, 1, ty], [0, 0, 1], **kwargs)


def translate3d(tx, ty, tz, **kwargs):
    """Construct 3d translation matrix.

    Args:
        tx (float): X-direction translation amount.
        ty (float): Y-direction translation amount.
        tz (float): Z-direction translation amount.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return matrix([1, 0, 0, tx], [0, 1, 0, ty], [0, 0, 1, tz], [0, 0, 0, 1],
                  **kwargs)


def scale2d(sx, sy, **kwargs):
    """Construct 2d scaling matrix.

    Args:
        sx (float): X-direction scaling coefficient.
        sy (float): Y-direction scaling coefficient.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return matrix([sx, 0, 0], [0, sy, 0], [0, 0, 1], **kwargs)


def scale3d(sx, sy, sz, **kwargs):
    """Construct 3d scaling matrix.

    Args:
        sx (float): X-direction scaling coefficient.
        sy (float): Y-direction scaling coefficient.
        sz (float): Z-direction scaling coefficient.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return matrix([sx, 0, 0, 0], [0, sy, 0, 0], [0, 0, sz, 0], [0, 0, 0, 1],
                  **kwargs)


def rotate2d(theta, **kwargs):
    """Construct 2d rotating matrix.

    Args:
        theta (float): Counter-clock wise rotation angle.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return matrix([torch.cos(theta), torch.sin(-theta), 0],
                  [torch.sin(theta), torch.cos(theta), 0], [0, 0, 1], **kwargs)


def rotate3d(v, theta, **kwargs):
    """Constructing 3d rotating matrix.

    Args:
        v (torch.Tensor): Luma axis.
        theta (float): Rotate theta counter-clock wise with ``v`` as the axis.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    vx = v[..., 0]
    vy = v[..., 1]
    vz = v[..., 2]
    s = torch.sin(theta)
    c = torch.cos(theta)
    cc = 1 - c
    return matrix(
        [vx * vx * cc + c, vx * vy * cc - vz * s, vx * vz * cc + vy * s, 0],
        [vy * vx * cc + vz * s, vy * vy * cc + c, vy * vz * cc - vx * s, 0],
        [vz * vx * cc - vy * s, vz * vy * cc + vx * s, vz * vz * cc + c, 0],
        [0, 0, 0, 1], **kwargs)


def translate2d_inv(tx, ty, **kwargs):
    """Construct inverse matrix of 2d translation matrix.

    Args:
        tx (float): X-direction translation amount.
        ty (float): Y-direction translation amount.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return translate2d(-tx, -ty, **kwargs)


def scale2d_inv(sx, sy, **kwargs):
    """Construct inverse matrix of 2d scaling matrix.

    Args:
        sx (float): X-direction scaling coefficient.
        sy (float): Y-direction scaling coefficient.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return scale2d(1 / sx, 1 / sy, **kwargs)


def rotate2d_inv(theta, **kwargs):
    """Construct inverse matrix of 2d rotating matrix.

    Args:
        theta (float): Counter-clock wise rotation angle.
    Returns:
        ndarry | Tensor : Transformation matrices in np.ndarry or torch.Tensor
            format.
    """
    return rotate2d(-theta, **kwargs)


# ----------------------------------------------------------------------------
# Versatile image augmentation pipeline from the paper
# "Training Generative Adversarial Networks with Limited Data".
#
# All augmentations are disabled by default; individual augmentations can
# be enabled by setting their probability multipliers to 1.


class AugmentPipe(torch.nn.Module):
    """Augmentation pipeline include multiple geometric and color
    transformations.

    Note: The meaning of arguments are written in the comments of
    ``__init__`` function.
    """

    def __init__(
        self,
        xflip=0,
        rotate90=0,
        xint=0,
        xint_max=0.125,
        scale=0,
        rotate=0,
        aniso=0,
        xfrac=0,
        scale_std=0.2,
        rotate_max=1,
        aniso_std=0.2,
        xfrac_std=0.125,
        brightness=0,
        contrast=0,
        lumaflip=0,
        hue=0,
        saturation=0,
        brightness_std=0.2,
        contrast_std=0.5,
        hue_max=1,
        saturation_std=1,
        imgfilter=0,
        imgfilter_bands=[1, 1, 1, 1],
        imgfilter_std=1,
        noise=0,
        cutout=0,
        noise_std=0.1,
        cutout_size=0.5,
    ):
        super().__init__()
        self.register_buffer('p', torch.ones(
            []))  # Overall multiplier for augmentation probability.

        # Pixel blitting.
        self.xflip = float(xflip)  # Probability multiplier for x-flip.
        self.rotate90 = float(
            rotate90)  # Probability multiplier for 90 degree rotations.
        self.xint = float(
            xint)  # Probability multiplier for integer translation.
        self.xint_max = float(
            xint_max
        )  # Range of integer translation, relative to image dimensions.

        # General geometric transformations.
        self.scale = float(
            scale)  # Probability multiplier for isotropic scaling.
        self.rotate = float(
            rotate)  # Probability multiplier for arbitrary rotation.
        self.aniso = float(
            aniso)  # Probability multiplier for anisotropic scaling.
        self.xfrac = float(
            xfrac)  # Probability multiplier for fractional translation.
        self.scale_std = float(
            scale_std)  # Log2 standard deviation of isotropic scaling.
        self.rotate_max = float(
            rotate_max)  # Range of arbitrary rotation, 1 = full circle.
        self.aniso_std = float(
            aniso_std)  # Log2 standard deviation of anisotropic scaling.
        self.xfrac_std = float(
            xfrac_std
        )  # Standard deviation of frational translation, relative to img dims.

        # Color transformations.
        self.brightness = float(
            brightness)  # Probability multiplier for brightness.
        self.contrast = float(contrast)  # Probability multiplier for contrast.
        self.lumaflip = float(
            lumaflip)  # Probability multiplier for luma flip.
        self.hue = float(hue)  # Probability multiplier for hue rotation.
        self.saturation = float(
            saturation)  # Probability multiplier for saturation.
        self.brightness_std = float(
            brightness_std)  # Standard deviation of brightness.
        self.contrast_std = float(
            contrast_std)  # Log2 standard deviation of contrast.
        self.hue_max = float(
            hue_max)  # Range of hue rotation, 1 = full circle.
        self.saturation_std = float(
            saturation_std)  # Log2 standard deviation of saturation.

        # Image-space filtering.
        self.imgfilter = float(
            imgfilter)  # Probability multiplier for image-space filtering.
        self.imgfilter_bands = list(
            imgfilter_bands
        )  # Probability multipliers for individual frequency bands.
        self.imgfilter_std = float(
            imgfilter_std
        )  # Log2 standard deviation of image-space filter amplification.

        # Image-space corruptions.
        self.noise = float(
            noise)  # Probability multiplier for additive RGB noise.
        self.cutout = float(cutout)  # Probability multiplier for cutout.
        self.noise_std = float(
            noise_std)  # Standard deviation of additive RGB noise.
        self.cutout_size = float(
            cutout_size
        )  # Size of the cutout rectangle, relative to image dimensions.

        # Setup orthogonal lowpass filter for geometric augmentations.
        self.register_buffer('Hz_geom',
                             upfirdn2d.setup_filter(wavelets['sym6']))

        # Construct filter bank for image-space filtering.
        Hz_lo = np.asarray(wavelets['sym2'])  # H(z)
        Hz_hi = Hz_lo * ((-1)**np.arange(Hz_lo.size))  # H(-z)
        Hz_lo2 = np.convolve(Hz_lo, Hz_lo[::-1]) / 2  # H(z) * H(z^-1) / 2
        Hz_hi2 = np.convolve(Hz_hi, Hz_hi[::-1]) / 2  # H(-z) * H(-z^-1) / 2
        Hz_fbank = np.eye(4, 1)  # Bandpass(H(z), b_i)
        for i in range(1, Hz_fbank.shape[0]):
            Hz_fbank = np.dstack([Hz_fbank, np.zeros_like(Hz_fbank)
                                  ]).reshape(Hz_fbank.shape[0], -1)[:, :-1]
            Hz_fbank = scipy.signal.convolve(Hz_fbank, [Hz_lo2])
            Hz_fbank[i, (Hz_fbank.shape[1] - Hz_hi2.size) //
                     2:(Hz_fbank.shape[1] + Hz_hi2.size) // 2] += Hz_hi2
        self.register_buffer('Hz_fbank',
                             torch.as_tensor(Hz_fbank, dtype=torch.float32))

    def forward(self, images, debug_percentile=None):
        assert isinstance(images, torch.Tensor) and images.ndim == 4
        batch_size, num_channels, height, width = images.shape
        device = images.device
        if debug_percentile is not None:
            debug_percentile = torch.as_tensor(
                debug_percentile, dtype=torch.float32, device=device)

        # -------------------------------------
        # Select parameters for pixel blitting.
        # -------------------------------------

        # Initialize inverse homogeneous 2D transform:
        # G_inv @ pixel_out ==> pixel_in
        I_3 = torch.eye(3, device=device)
        G_inv = I_3

        # Apply x-flip with probability (xflip * strength).
        if self.xflip > 0:
            i = torch.floor(torch.rand([batch_size], device=device) * 2)
            i = torch.where(
                torch.rand([batch_size], device=device) < self.xflip * self.p,
                i, torch.zeros_like(i))
            if debug_percentile is not None:
                i = torch.full_like(i, torch.floor(debug_percentile * 2))
            G_inv = G_inv @ scale2d_inv(1 - 2 * i, 1)

        # Apply 90 degree rotations with probability (rotate90 * strength).
        if self.rotate90 > 0:
            i = torch.floor(torch.rand([batch_size], device=device) * 4)
            i = torch.where(
                torch.rand([batch_size], device=device) <
                self.rotate90 * self.p, i, torch.zeros_like(i))
            if debug_percentile is not None:
                i = torch.full_like(i, torch.floor(debug_percentile * 4))
            G_inv = G_inv @ rotate2d_inv(-np.pi / 2 * i)

        # Apply integer translation with probability (xint * strength).
        if self.xint > 0:
            t = (torch.rand([batch_size, 2], device=device) * 2 -
                 1) * self.xint_max
            t = torch.where(
                torch.rand([batch_size, 1], device=device) <
                self.xint * self.p, t, torch.zeros_like(t))
            if debug_percentile is not None:
                t = torch.full_like(t,
                                    (debug_percentile * 2 - 1) * self.xint_max)
            G_inv = G_inv @ translate2d_inv(
                torch.round(t[:, 0] * width), torch.round(t[:, 1] * height))

        # --------------------------------------------------------
        # Select parameters for general geometric transformations.
        # --------------------------------------------------------

        # support for pt1.5 (pt1.5 does not contain exp2)
        _scalor_log2 = torch.log(
            torch.tensor(2., device=images.device, dtype=images.dtype))

        # Apply isotropic scaling with probability (scale * strength).
        if self.scale > 0:
            s = torch.exp(
                torch.randn([batch_size], device=device) * self.scale_std *
                _scalor_log2)
            s = torch.where(
                torch.rand([batch_size], device=device) < self.scale * self.p,
                s, torch.ones_like(s))
            if debug_percentile is not None:
                s = torch.full_like(
                    s,
                    torch.exp2(
                        torch.erfinv(debug_percentile * 2 - 1) *
                        self.scale_std))
            G_inv = G_inv @ scale2d_inv(s, s)

        # Apply pre-rotation with probability p_rot.
        p_rot = 1 - torch.sqrt(
            (1 - self.rotate * self.p).clamp(0, 1))  # P(pre OR post) = p
        if self.rotate > 0:
            theta = (torch.rand([batch_size], device=device) * 2 -
                     1) * np.pi * self.rotate_max
            theta = torch.where(
                torch.rand([batch_size], device=device) < p_rot, theta,
                torch.zeros_like(theta))
            if debug_percentile is not None:
                theta = torch.full_like(theta, (debug_percentile * 2 - 1) *
                                        np.pi * self.rotate_max)
            G_inv = G_inv @ rotate2d_inv(-theta)  # Before anisotropic scaling.

        # Apply anisotropic scaling with probability (aniso * strength).
        if self.aniso > 0:
            s = torch.exp(
                torch.randn([batch_size], device=device) * self.aniso_std *
                _scalor_log2)
            s = torch.where(
                torch.rand([batch_size], device=device) < self.aniso * self.p,
                s, torch.ones_like(s))
            if debug_percentile is not None:
                s = torch.full_like(
                    s,
                    torch.exp2(
                        torch.erfinv(debug_percentile * 2 - 1) *
                        self.aniso_std))
            G_inv = G_inv @ scale2d_inv(s, 1 / s)

        # Apply post-rotation with probability p_rot.
        if self.rotate > 0:
            theta = (torch.rand([batch_size], device=device) * 2 -
                     1) * np.pi * self.rotate_max
            theta = torch.where(
                torch.rand([batch_size], device=device) < p_rot, theta,
                torch.zeros_like(theta))
            if debug_percentile is not None:
                theta = torch.zeros_like(theta)
            G_inv = G_inv @ rotate2d_inv(-theta)  # After anisotropic scaling.

        # Apply fractional translation with probability (xfrac * strength).
        if self.xfrac > 0:
            t = torch.randn([batch_size, 2], device=device) * self.xfrac_std
            t = torch.where(
                torch.rand([batch_size, 1], device=device) <
                self.xfrac * self.p, t, torch.zeros_like(t))
            if debug_percentile is not None:
                t = torch.full_like(
                    t,
                    torch.erfinv(debug_percentile * 2 - 1) * self.xfrac_std)
            G_inv = G_inv @ translate2d_inv(t[:, 0] * width, t[:, 1] * height)

        # ----------------------------------
        # Execute geometric transformations.
        # ----------------------------------

        # Execute if the transform is not identity.
        if G_inv is not I_3:

            # Calculate padding.
            cx = (width - 1) / 2
            cy = (height - 1) / 2
            cp = matrix([-cx, -cy, 1], [cx, -cy, 1], [cx, cy, 1], [-cx, cy, 1],
                        device=device)  # [idx, xyz]
            cp = G_inv @ cp.t()  # [batch, xyz, idx]
            Hz_pad = self.Hz_geom.shape[0] // 4
            margin = cp[:, :2, :].permute(1, 0,
                                          2).flatten(1)  # [xy, batch * idx]
            margin = torch.cat([-margin,
                                margin]).max(dim=1).values  # [x0, y0, x1, y1]
            margin = margin + misc.constant(
                [Hz_pad * 2 - cx, Hz_pad * 2 - cy] * 2, device=device)
            margin = margin.max(misc.constant([0, 0] * 2, device=device))
            margin = margin.min(
                misc.constant([width - 1, height - 1] * 2, device=device))
            mx0, my0, mx1, my1 = margin.ceil().to(torch.int32)

            # Pad image and adjust origin.
            images = torch.nn.functional.pad(
                input=images, pad=[mx0, mx1, my0, my1], mode='reflect')
            G_inv = translate2d(
                torch.true_divide(mx0 - mx1, 2), torch.true_divide(
                    my0 - my1, 2)) @ G_inv

            # Upsample.
            images = upfirdn2d.upsample2d(x=images, f=self.Hz_geom, up=2)
            G_inv = scale2d(
                2, 2, device=device) @ G_inv @ scale2d_inv(
                    2, 2, device=device)
            G_inv = translate2d(
                -0.5, -0.5, device=device) @ G_inv @ translate2d_inv(
                    -0.5, -0.5, device=device)

            # Execute transformation.
            shape = [
                batch_size, num_channels, (height + Hz_pad * 2) * 2,
                (width + Hz_pad * 2) * 2
            ]
            G_inv = scale2d(
                2 / images.shape[3], 2 / images.shape[2],
                device=device) @ G_inv @ scale2d_inv(
                    2 / shape[3], 2 / shape[2], device=device)
            grid = torch.nn.functional.affine_grid(
                theta=G_inv[:, :2, :], size=shape, align_corners=False)
            images = grid_sample_gradfix.grid_sample(images, grid)

            # Downsample and crop.
            images = upfirdn2d.downsample2d(
                x=images,
                f=self.Hz_geom,
                down=2,
                padding=-Hz_pad * 2,
                flip_filter=True)

        # --------------------------------------------
        # Select parameters for color transformations.
        # --------------------------------------------

        # Initialize homogeneous 3D transformation matrix:
        # C @ color_in ==> color_out
        I_4 = torch.eye(4, device=device)
        C = I_4

        # Apply brightness with probability (brightness * strength).
        if self.brightness > 0:
            b = torch.randn([batch_size], device=device) * self.brightness_std
            b = torch.where(
                torch.rand([batch_size], device=device) <
                self.brightness * self.p, b, torch.zeros_like(b))
            if debug_percentile is not None:
                b = torch.full_like(
                    b,
                    torch.erfinv(debug_percentile * 2 - 1) *
                    self.brightness_std)
            C = translate3d(b, b, b) @ C

        # Apply contrast with probability (contrast * strength).
        if self.contrast > 0:
            c = torch.exp(
                torch.randn([batch_size], device=device) * self.contrast_std *
                _scalor_log2)
            c = torch.where(
                torch.rand([batch_size], device=device) <
                self.contrast * self.p, c, torch.ones_like(c))
            if debug_percentile is not None:
                c = torch.full_like(
                    c,
                    torch.exp2(
                        torch.erfinv(debug_percentile * 2 - 1) *
                        self.contrast_std))
            C = scale3d(c, c, c) @ C

        # Apply luma flip with probability (lumaflip * strength).
        v = misc.constant(
            np.asarray([1, 1, 1, 0]) / np.sqrt(3), device=device)  # Luma axis.
        if self.lumaflip > 0:
            i = torch.floor(torch.rand([batch_size, 1, 1], device=device) * 2)
            i = torch.where(
                torch.rand([batch_size, 1, 1], device=device) <
                self.lumaflip * self.p, i, torch.zeros_like(i))
            if debug_percentile is not None:
                i = torch.full_like(i, torch.floor(debug_percentile * 2))
            C = (I_4 - 2 * v.ger(v) * i) @ C  # Householder reflection.

        # Apply hue rotation with probability (hue * strength).
        if self.hue > 0 and num_channels > 1:
            theta = (torch.rand([batch_size], device=device) * 2 -
                     1) * np.pi * self.hue_max
            theta = torch.where(
                torch.rand([batch_size], device=device) < self.hue * self.p,
                theta, torch.zeros_like(theta))
            if debug_percentile is not None:
                theta = torch.full_like(theta, (debug_percentile * 2 - 1) *
                                        np.pi * self.hue_max)
            C = rotate3d(v, theta) @ C  # Rotate around v.

        # Apply saturation with probability (saturation * strength).
        if self.saturation > 0 and num_channels > 1:
            s = torch.exp(
                torch.randn([batch_size, 1, 1], device=device) *
                self.saturation_std * _scalor_log2)
            s = torch.where(
                torch.rand([batch_size, 1, 1], device=device) <
                self.saturation * self.p, s, torch.ones_like(s))
            if debug_percentile is not None:
                s = torch.full_like(
                    s,
                    torch.exp2(
                        torch.erfinv(debug_percentile * 2 - 1) *
                        self.saturation_std))
            C = (v.ger(v) + (I_4 - v.ger(v)) * s) @ C

        # ------------------------------
        # Execute color transformations.
        # ------------------------------

        # Execute if the transform is not identity.
        if C is not I_4:
            images = images.reshape([batch_size, num_channels, height * width])
            if num_channels == 3:
                images = C[:, :3, :3] @ images + C[:, :3, 3:]
            elif num_channels == 1:
                C = C[:, :3, :].mean(dim=1, keepdims=True)
                images = images * C[:, :, :3].sum(
                    dim=2, keepdims=True) + C[:, :, 3:]
            else:
                raise ValueError(
                    'Image must be RGB (3 channels) or L (1 channel)')
            images = images.reshape([batch_size, num_channels, height, width])

        # ----------------------
        # Image-space filtering.
        # ----------------------

        if self.imgfilter > 0:
            num_bands = self.Hz_fbank.shape[0]
            assert len(self.imgfilter_bands) == num_bands
            expected_power = misc.constant(
                np.array([10, 1, 1, 1]) / 13,
                device=device)  # Expected power spectrum (1/f).

            # Apply amplification for each band with probability
            # (imgfilter * strength * band_strength).
            g = torch.ones([batch_size, num_bands],
                           device=device)  # Global gain vector (identity).
            for i, band_strength in enumerate(self.imgfilter_bands):
                t_i = torch.exp(
                    torch.randn([batch_size], device=device) *
                    self.imgfilter_std * _scalor_log2)
                t_i = torch.where(
                    torch.rand([batch_size], device=device) <
                    self.imgfilter * self.p * band_strength, t_i,
                    torch.ones_like(t_i))
                if debug_percentile is not None:
                    t_i = torch.full_like(
                        t_i,
                        torch.exp2(
                            torch.erfinv(debug_percentile * 2 - 1) *
                            self.imgfilter_std)
                    ) if band_strength > 0 else torch.ones_like(t_i)
                t = torch.ones([batch_size, num_bands],
                               device=device)  # Temporary gain vector.
                t[:, i] = t_i  # Replace i'th element.
                t = t / (expected_power * t.square()).sum(
                    dim=-1, keepdims=True).sqrt()  # Normalize power.
                g = g * t  # Accumulate into global gain.

            # Construct combined amplification filter.
            Hz_prime = g @ self.Hz_fbank  # [batch, tap]
            Hz_prime = Hz_prime.unsqueeze(1).repeat(
                [1, num_channels, 1])  # [batch, channels, tap]
            Hz_prime = Hz_prime.reshape([batch_size * num_channels, 1,
                                         -1])  # [batch * channels, 1, tap]

            # Apply filter.
            p = self.Hz_fbank.shape[1] // 2
            images = images.reshape(
                [1, batch_size * num_channels, height, width])
            images = torch.nn.functional.pad(
                input=images, pad=[p, p, p, p], mode='reflect')
            images = conv2d(
                input=images,
                weight=Hz_prime.unsqueeze(2),
                groups=batch_size * num_channels)
            images = conv2d(
                input=images,
                weight=Hz_prime.unsqueeze(3),
                groups=batch_size * num_channels)
            images = images.reshape([batch_size, num_channels, height, width])

        # ------------------------
        # Image-space corruptions.
        # ------------------------

        # Apply additive RGB noise with probability (noise * strength).
        if self.noise > 0:
            sigma = torch.randn([batch_size, 1, 1, 1],
                                device=device).abs() * self.noise_std
            sigma = torch.where(
                torch.rand([batch_size, 1, 1, 1], device=device) <
                self.noise * self.p, sigma, torch.zeros_like(sigma))
            if debug_percentile is not None:
                sigma = torch.full_like(
                    sigma,
                    torch.erfinv(debug_percentile) * self.noise_std)
            images = images + torch.randn(
                [batch_size, num_channels, height, width],
                device=device) * sigma

        # Apply cutout with probability (cutout * strength).
        if self.cutout > 0:
            size = torch.full([batch_size, 2, 1, 1, 1],
                              self.cutout_size,
                              device=device)
            size = torch.where(
                torch.rand([batch_size, 1, 1, 1, 1], device=device) <
                self.cutout * self.p, size, torch.zeros_like(size))
            center = torch.rand([batch_size, 2, 1, 1, 1], device=device)
            if debug_percentile is not None:
                size = torch.full_like(size, self.cutout_size)
                center = torch.full_like(center, debug_percentile)
            coord_x = torch.arange(width, device=device).reshape([1, 1, 1, -1])
            coord_y = torch.arange(
                height, device=device).reshape([1, 1, -1, 1])
            mask_x = (((coord_x + 0.5) / width - center[:, 0]).abs() >=
                      size[:, 0] / 2)
            mask_y = (((coord_y + 0.5) / height - center[:, 1]).abs() >=
                      size[:, 1] / 2)
            mask = torch.logical_or(mask_x, mask_y).to(torch.float32)
            images = images * mask

        return images
